import logging
from typing import Any, AsyncGenerator, Dict, List, Optional, Union

from nawah.classes import ATTR, NAWAH_ENV, Query
from nawah.config import Config

from ._query import _compile_query
from ._read import _process_results_doc

logger = logging.getLogger('nawah')


async def watch(
    *,
    env: NAWAH_ENV,
    collection_name: str,
    attrs: Dict[str, ATTR],
    query: Query,
    skip_extn: bool = False,
) -> AsyncGenerator[Dict[str, Any], Dict[str, Any]]:
    aggregate_query = _compile_query(
        collection_name=collection_name, attrs=attrs, query=query, watch_mode=True
    )[4]

    collection = env['conn'][Config.data_name][collection_name]

    logger.debug('Preparing generator at Data')
    async with collection.watch(
        pipeline=aggregate_query, full_document='updateLookup'
    ) as stream:
        yield {'stream': stream}
        async for change in stream:
            logger.debug(f'Detected change at Data: {change}')

            oper = change['operationType']
            if oper in ['insert', 'replace', 'update']:
                if oper == 'insert':
                    oper = 'create'
                elif oper == 'replace':
                    oper = 'update'
                doc = await _process_results_doc(
                    env=env,
                    collection=collection,
                    attrs=attrs,
                    doc=change['fullDocument'],
                    skip_extn=skip_extn,
                )
                model = doc
            elif oper == 'delete':
                model = {'_id': change['documentKey']['_id']}

            yield {'count': 1, 'oper': oper, 'docs': [model]}

    logger.debug('changeStream has been close. Generator ended at Data')



'''Exposes package members to be used by other packages'''

from ._attr import (_deep_update, _expand_attr, _extract_attr, _set_attr,
                    _update_attr_values)
from ._config import (_compile_anon_session, _compile_anon_user, _config_data,
                      _process_config)
from ._encode_attr_type import encode_attr_type
from ._generate_attr import generate_attr
from ._generate_models import _generate_models
from ._generate_ref import _extract_lambda_body, _generate_ref
from ._import_modules import _import_modules
from ._reset_cache import reset_cache
from ._validate import (_process_file_obj, _set_none_value,
                        generate_dynamic_attr, validate_attr, validate_doc)



from nawah.base_module import BaseModule
from nawah.classes import ATTR, METHOD, PERM
from nawah.config import Config
from nawah.enums import Event


class Group(BaseModule):
    '''`Group` module provides data type and controller for groups in Nawah eco-system.'''

    collection = 'groups'
    attrs = {
        'user': ATTR.ID(desc='`_id` of `User` doc the doc belongs to.'),
        'name': ATTR.LOCALE(desc='Name of the groups as `LOCALE`.'),
        'desc': ATTR.LOCALE(
            desc='Description of the group as `LOCALE`. This can be used for dynamic generated groups that are meant to be exposed to end-users.'
        ),
        'privileges': ATTR.KV_DICT(
            desc='Privileges that any user is a member of the group has.',
            key=ATTR.STR(),
            val=ATTR.LIST(list=[ATTR.STR()]),
        ),
        'settings': ATTR.KV_DICT(
            desc='`Setting` docs to be created, or required for members users when added to the group.',
            key=ATTR.STR(),
            val=ATTR.ANY(),
        ),
        'create_time': ATTR.DATETIME(
            desc='Python `datetime` ISO format of the doc creation.'
        ),
    }
    defaults = {
        'desc': {locale: '' for locale in Config.locales},
        'privileges': {},
        'settings': {},
    }
    methods = {
        'read': METHOD(permissions=[PERM(privilege='admin')]),
        'create': METHOD(permissions=[PERM(privilege='admin')]),
        'update': METHOD(
            permissions=[
                PERM(privilege='admin'),
                PERM(
                    privilege='update',
                    query_mod={'user': '$__user'},
                    doc_mod={'privileges': None},
                ),
            ],
            query_args={'_id': ATTR.ID()},
        ),
        'delete': METHOD(
            permissions=[
                PERM(privilege='admin'),
                PERM(privilege='delete', query_mod={'user': '$__user'}),
            ],
            query_args={'_id': ATTR.ID()},
        ),
    }

    async def pre_update(self, skip_events, env, query, doc, payload):
        # [DOC] Make sure no attrs overwriting would happen
        if 'attrs' in doc.keys():
            results = await self.read(skip_events=[Event.PERM], env=env, query=query)
            if not results['args']['count']:
                raise self.exception(
                    status=400, msg='Group is invalid.', args={'code': 'INVALID_GROUP'}
                )

            if results['args']['count'] > 1:
                raise self.exception(
                    status=400,
                    msg='Updating group attrs can be done only to individual groups.',
                    args={'code': 'MULTI_ATTRS_UPDATE'},
                )

            results['args']['docs'][0]['attrs'].update(
                {
                    attr: doc['attrs'][attr]
                    for attr in doc['attrs'].keys()
                    if doc['attrs'][attr] != None and doc['attrs'][attr] != ''
                }
            )
            doc['attrs'] = results['args']['docs'][0]['attrs']
        return (skip_events, env, query, doc, payload)



from typing import Literal, Any

import logging

logger = logging.getLogger('nawah')


class InvalidPermissionsExcpetion(Exception):
    pass


class InvalidCallArgsException(Exception):
    pass


class InvalidAttrTypeException(Exception):
    def __init__(self, *, attr_type: Any):
        self.attr_type = attr_type

    def __str__(self):
        return f'Unknown or invalid Attr Type \'{self.attr_type}\'.'


class InvalidAttrTypeArgException(Exception):
    def __init__(self, *, arg_name: str, arg_type: Any, arg_val: Any):
        self.arg_name = arg_name
        self.arg_type = arg_type
        self.arg_val = arg_val

    def __str__(self):
        return f'Invalid Attr Type Arg for \'{self.arg_name}\' expecting type \'{self.arg_type}\' but got \'{self.arg_val}\'.'


class InvalidAttrTypeArgsException(Exception):
    def __init__(self, *, msg: str):
        self.msg = msg

    def __str__(self):
        return self.msg


class MethodException(Exception):
    pass


class InvalidQueryArgException(Exception):
    def __init__(
        self,
        *,
        arg_name: str,
        arg_oper: Literal[
            '$ne',
            '$eq',
            '$gt',
            '$gte',
            '$lt',
            '$lte',
            '$bet',
            '$all',
            '$in',
            '$nin',
            '$regex',
        ],
        arg_type: Any,
        arg_val: Any,
    ):
        self.arg_name = arg_name
        self.arg_oper = arg_oper
        self.arg_type = arg_type
        self.arg_val = arg_val

    def __str__(self):
        return f'Invalid value for Query Arg \'{self.arg_name}\' with Query Arg Oper \'{self.arg_oper}\' expecting type \'{self.arg_type}\' but got \'{self.arg_val}\'.'


class UnknownQueryArgException(Exception):
    def __init__(
        self,
        *,
        arg_name: str,
        arg_oper: Literal[
            '$ne',
            '$eq',
            '$gt',
            '$gte',
            '$lt',
            '$lte',
            '$bet',
            '$all',
            '$in',
            '$nin',
            '$regex',
        ],
    ):
        self.arg_name = arg_name
        self.arg_oper = arg_oper

    def __str__(self):
        return f'Unknown Query Arg Oper \'{self.arg_oper}\' for Query Arg \'{self.arg_name}\'.'


class MissingAttrException(Exception):
    def __init__(self, *, attr_name):
        self.attr_name = attr_name
        logger.debug(f'MissingAttrException: {str(self)}')

    def __str__(self):
        return f'Missing attr \'{self.attr_name}\''


class InvalidAttrException(Exception):
    def __init__(self, *, attr_name, attr_type, val_type):
        self.attr_name = attr_name
        self.attr_type = attr_type
        self.val_type = val_type
        logger.debug(f'InvalidAttrException: {str(self)}')

    def __str__(self):
        return f'Invalid attr \'{self.attr_name}\' of type \'{self.val_type}\' with required type \'{self.attr_type}\''


class ConvertAttrException(Exception):
    def __init__(self, *, attr_name, attr_type, val_type):
        self.attr_name = attr_name
        self.attr_type = attr_type
        self.val_type = val_type
        logger.debug(f'ConvertAttrException: {str(self)}')

    def __str__(self):
        return f'Can\'t convert attr \'{self.attr_name}\' of type \'{self.val_type}\' to type \'{self.attr_type._type}\''


class InvalidModuleException(Exception):
    def __init__(self, *, module):
        self.module = module
        logger.debug(f'InvalidModuleException: {str(self)}')

    def __str__(self):
        return f'Invalid module \'{self.module}\''


class InvalidLocaleException(Exception):
    def __init__(self, *, locale):
        self.locale = locale
        logger.debug(f'InvalidLocaleException: {str(self)}')

    def __str__(self):
        return f'Invalid locale \'{self.locale}\''


class InvalidLocaleTermException(Exception):
    def __init__(self, *, locale, term):
        self.locale = locale
        self.term = term
        logger.debug(f'InvalidLocaleTermException: {str(self)}')

    def __str__(self):
        return f'Invalid term \'{self.term}\' of locale \'{self.locale}\''


class InvalidVarException(Exception):
    def __init__(self, *, var):
        self.var = var
        logger.debug(f'InvalidVarException: {str(self)}')

    def __str__(self):
        return f'Invalid var \'{self.var}\''


class UnknownDeleteStrategyException(Exception):
    pass


class InvalidQueryException(Exception):
    pass


class InvalidGatewayException(Exception):
    def __init__(self, *, gateway):
        self.gateway = gateway

    def __str__(self):
        return f'Gateway \'{self.gateway}\' is invalid.'


class UnexpectedGatewayException(Exception):
    def __init__(self, *, gateway):
        self.gateway = gateway

    def __str__(self):
        return f'An unexpected gateway exception occurred when attempted to call \'{self.gateway}\'.'



from typing import Dict, List, cast


async def run_app():
    import asyncio
    import datetime
    import json
    import logging
    import time
    import traceback

    import aiohttp.web
    import jwt
    import nest_asyncio
    from multidict import MultiDict
    from passlib.hash import pbkdf2_sha512
    from requests_toolbelt.multipart import decoder

    from nawah import data as Data
    from nawah.classes import (ATTR, IP_QUOTA, NAWAH_ENV, ConvertAttrException,
                               InvalidAttrException, JSONEncoder)
    from nawah.config import Config
    from nawah.enums import Event
    from nawah.utils import (_compile_anon_session, _compile_anon_user,
                             _config_data, _import_modules, _process_file_obj,
                             validate_doc)

    nest_asyncio.apply()

    logger = logging.getLogger('nawah')

    # [DOC] Use import_modules to load and initialise modules
    await _import_modules()
    await _config_data()
    # [DOC] Populate get_routes, post_routes
    get_routes = []
    post_routes = []
    for module in Config.modules.values():
        for method in module.methods.values():
            if method.get_method:
                for get_args_set in method.query_args:
                    if get_args_set:
                        get_args = f'/{{{"}/{".join(list(get_args_set.keys()))}}}'
                    else:
                        get_args = ''

                    get_routes.append(
                        f'/{module.module_name}/{method._callable.method}{get_args}'
                    )
            elif method.post_method:
                for post_args_set in method.query_args:
                    if post_args_set:
                        post_args = f'/{{{"}/{".join(list(post_args_set.keys()))}}}'
                    else:
                        post_args = ''

                    post_routes.append(
                        f'/{module.module_name}/{method._callable.method}{post_args}'
                    )

    logger.debug(
        'Loaded modules: %s',
        {module: Config.modules[module].attrs for module in Config.modules.keys()},
    )
    logger.debug(
        'Config has attrs: %s',
        {
            k: str(v)
            for k, v in Config.__dict__.items()
            if not type(v) == classmethod and not k.startswith('_')
        },
    )
    logger.debug(f'Generated get_routes: {get_routes}')
    logger.debug(f'Generated post_routes: {post_routes}')

    sessions: List[NAWAH_ENV] = []
    ip_quota: Dict[str, IP_QUOTA] = {}

    async def not_found_handler(request):
        headers = MultiDict(
            [
                ('Server', 'Nawah'),
                ('Powered-By', 'Nawah, https://nawah.masaar.com'),
                ('Access-Control-Allow-Origin', '*'),
                ('Access-Control-Allow-Methods', 'GET,POST'),
                ('Access-Control-Allow-Headers', 'Content-Type'),
                ('Access-Control-Expose-Headers', 'Content-Disposition'),
            ]
        )
        return aiohttp.web.Response(
            status=404,
            headers=headers,
            body=JSONEncoder().encode({'status': 404, 'msg': '404 NOT FOUND'}),
        )

    async def not_allowed_handler(request):
        headers = MultiDict(
            [
                ('Server', 'Nawah'),
                ('Powered-By', 'Nawah, https://nawah.masaar.com'),
                ('Access-Control-Allow-Origin', '*'),
                ('Access-Control-Allow-Methods', '*'),
                ('Access-Control-Allow-Headers', 'Content-Type'),
                ('Access-Control-Expose-Headers', 'Content-Disposition'),
            ]
        )
        return aiohttp.web.Response(
            status=405,
            headers=headers,
            body=JSONEncoder().encode({'status': 405, 'msg': '404 NOT ALLOWED'}),
        )

    async def root_handler(request: aiohttp.web.Request):
        headers = MultiDict(
            [
                ('Server', 'Nawah'),
                ('Powered-By', 'Nawah, https://nawah.masaar.com'),
                ('Access-Control-Allow-Origin', '*'),
                ('Access-Control-Allow-Methods', 'GET'),
                ('Access-Control-Allow-Headers', 'Content-Type'),
                ('Access-Control-Expose-Headers', 'Content-Disposition'),
            ]
        )
        return aiohttp.web.Response(
            status=200,
            headers=headers,
            body=JSONEncoder().encode(
                {
                    'status': 200,
                    'msg': f'Welcome to {Config._app_name}!',
                    'args': {'version': Config._app_version, 'powered_by': 'Nawah'},
                }
            ),
        )

    async def http_handler(request: aiohttp.web.Request):
        headers = MultiDict(
            [
                ('Server', 'Nawah'),
                ('Powered-By', 'Nawah, https://nawah.masaar.com'),
                ('Access-Control-Allow-Origin', '*'),
                ('Access-Control-Allow-Methods', 'GET,POST,OPTIONS'),
                (
                    'Access-Control-Allow-Headers',
                    'Content-Type,X-Auth-Bearer,X-Auth-Token,X-Auth-App',
                ),
                ('Access-Control-Expose-Headers', 'Content-Disposition'),
            ]
        )

        logger.debug(f'Received new {request.method} request: {request.match_info}')

        if request.method == 'OPTIONS':
            return aiohttp.web.Response(
                status=200,
                headers=headers,
                body=JSONEncoder().encode(
                    {
                        'status': 200,
                        'msg': 'OPTIONS request is allowed.',
                    }
                ),
            )

        # [DOC] Check for IP quota
        if str(request.remote) not in ip_quota:
            ip_quota[str(request.remote)] = {
                'counter': Config.quota_ip_min,
                'last_check': datetime.datetime.utcnow(),
            }
        else:
            if (
                datetime.datetime.utcnow() - ip_quota[str(request.remote)]['last_check']
            ).seconds > 259:
                ip_quota[str(request.remote)]['last_check'] = datetime.datetime.utcnow()
                ip_quota[str(request.remote)]['counter'] = Config.quota_ip_min
            else:
                if ip_quota[str(request.remote)]['counter'] - 1 <= 0:
                    logger.warning(
                        f'Denying \'{request.method}\' request from \'{request.remote}\' for hitting IP quota.'
                    )
                    headers['Content-Type'] = 'application/json; charset=utf-8'
                    return aiohttp.web.Response(
                        status=429,
                        headers=headers,
                        body=JSONEncoder().encode(
                            {
                                'status': 429,
                                'msg': 'You have hit calls quota from this IP.',
                                'args': {'code': 'CORE_REQ_IP_QUOTA_HIT'},
                            }
                        ),
                    )
                else:
                    ip_quota[str(request.remote)]['counter'] -= 1

        module = request.url.parts[1].lower()
        method = request.url.parts[2].lower()
        request_args = dict(request.match_info.items())

        # [DOC] Extract Args Sets based on request.method
        args_sets = Config.modules[module].methods[method].query_args
        args_sets = cast(List[Dict[str, ATTR]], args_sets)

        # [DOC] Attempt to validate query as doc
        for args_set in args_sets:
            if len(args_set.keys()) == len(args_set.keys()) and sum(
                1 for arg in args_set.keys() if arg in args_set.keys()
            ) == len(args_set.keys()):
                # [DOC] Check presence and validate all attrs in doc args
                try:
                    exception: Exception
                    await validate_doc(mode='create', doc=request_args, attrs=args_set)  # type: ignore
                except InvalidAttrException as e:
                    exception = e
                    headers['Content-Type'] = 'application/json; charset=utf-8'
                    return aiohttp.web.Response(
                        status=400,
                        headers=headers,
                        body=JSONEncoder()
                        .encode(
                            {
                                'status': 400,
                                'msg': f'{str(e)} for \'{request.method}\' request on module \'{Config.modules[module].package_name.upper()}_{module.upper()}\'.',
                                'args': {
                                    'code': f'{Config.modules[module].package_name.upper()}_{module.upper()}_INVALID_ATTR'
                                },
                            }
                        )
                        .encode('utf-8'),
                    )
                except ConvertAttrException as e:
                    exception = e
                    headers['Content-Type'] = 'application/json; charset=utf-8'
                    return aiohttp.web.Response(
                        status=400,
                        headers=headers,
                        body=JSONEncoder()
                        .encode(
                            {
                                'status': 400,
                                'msg': f'{str(e)} for \'{request.method}\' request on module \'{Config.modules[module].package_name.upper()}_{module.upper()}\'.',
                                'args': {
                                    'code': f'{Config.modules[module].package_name.upper()}_{module.upper()}_CONVERT_INVALID_ATTR'
                                },
                            }
                        )
                        .encode('utf-8'),
                    )
                break

        conn = Data.create_conn()
        env: NAWAH_ENV = {
            'conn': conn,
            'REMOTE_ADDR': request.remote,
            'client_app': '__public',
        }

        try:
            env['HTTP_USER_AGENT'] = request.headers['user-agent']
            env['HTTP_ORIGIN'] = request.headers['origin']
        except:
            env['HTTP_USER_AGENT'] = ''
            env['HTTP_ORIGIN'] = ''

        if 'X-Auth-Bearer' in request.headers or 'X-Auth-Token' in request.headers:
            logger.debug('Detected \'X-Auth\' header[s].')
            if (
                'X-Auth-Bearer' not in request.headers
                or 'X-Auth-Token' not in request.headers
                or 'X-Auth-App' not in request.headers
            ):
                logger.debug('Denying request due to missing \'X-Auth\' header.')
                headers['Content-Type'] = 'application/json; charset=utf-8'
                return aiohttp.web.Response(
                    status=400,
                    headers=headers,
                    body=JSONEncoder()
                    .encode(
                        {
                            'status': 400,
                            'msg': 'One \'X-Auth\' headers was set but not the other.',
                        }
                    )
                    .encode('utf-8'),
                )
            # [TODO] Add condition to now allow a connection from a client with client_app = __sys
            if len(Config.client_apps.keys()) and (
                request.headers['X-Auth-App'] not in Config.client_apps.keys()
                or (
                    Config.client_apps[request.headers['X-Auth-App']]['type'] == 'web'
                    and env['HTTP_ORIGIN']
                    not in Config.client_apps[request.headers['X-Auth-App']]['origin']
                )
            ):
                logger.debug('Denying request due to unauthorised client_app.')
                headers['Content-Type'] = 'application/json; charset=utf-8'
                return aiohttp.web.Response(
                    status=403,
                    headers=headers,
                    body=JSONEncoder()
                    .encode(
                        {
                            'status': 403,
                            'msg': 'X-Auth headers could not be verified.',
                            'args': {'code': 'CORE_SESSION_INVALID_XAUTH'},
                        }
                    )
                    .encode('utf-8'),
                )
            try:
                session_results = await Config.modules['session'].read(
                    skip_events=[Event.PERM],
                    env=env,
                    query=[
                        {
                            '_id': request.headers['X-Auth-Bearer'],
                        }
                    ],
                )
            except:
                headers['Content-Type'] = 'application/json; charset=utf-8'
                if Config.debug:
                    return aiohttp.web.Response(
                        status=500,
                        headers=headers,
                        body=JSONEncoder()
                        .encode(
                            {
                                'status': 500,
                                'msg': f'Unexpected error has occurred [{str(exception)}].',
                                'args': {
                                    'code': 'CORE_SERVER_ERROR',
                                    'err': str(exception),
                                },
                            }
                        )
                        .encode('utf-8'),
                    )
                else:
                    return aiohttp.web.Response(
                        status=500,
                        headers=headers,
                        body=JSONEncoder()
                        .encode(
                            {
                                'status': 500,
                                'msg': 'Unexpected error has occurred.',
                                'args': {'code': 'CORE_SERVER_ERROR'},
                            }
                        )
                        .encode('utf-8'),
                    )

            if not session_results['args']['count'] or not pbkdf2_sha512.verify(
                request.headers['X-Auth-Token'],
                session_results['args']['docs'][0]['token_hash'],
            ):
                logger.debug(
                    'Denying request due to missing failed Call Authorisation.'
                )
                headers['Content-Type'] = 'application/json; charset=utf-8'
                return aiohttp.web.Response(
                    status=403,
                    headers=headers,
                    body=JSONEncoder()
                    .encode(
                        {
                            'status': 403,
                            'msg': 'X-Auth headers could not be verified.',
                            'args': {'code': 'CORE_SESSION_INVALID_XAUTH'},
                        }
                    )
                    .encode('utf-8'),
                )
            else:
                session = session_results['args']['docs'][0]
                session_results = await Config.modules['session'].reauth(
                    skip_events=[Event.PERM],
                    env=env,
                    query=[
                        {
                            '_id': request.headers['X-Auth-Bearer'],
                            'token': request.headers['X-Auth-Token'],
                        }
                    ],
                )
                logger.debug('Denying request due to fail to reauth.')
                if session_results['status'] != 200:
                    headers['Content-Type'] = 'application/json; charset=utf-8'
                    return aiohttp.web.Response(
                        status=403,
                        headers=headers,
                        body=JSONEncoder().encode(session_results).encode('utf-8'),
                    )
                else:
                    session = session_results['args']['session']
        else:
            anon_user = _compile_anon_user()
            anon_session = _compile_anon_session()
            anon_session['user'] = anon_user
            session = anon_session

        env['session'] = session

        doc_content = await request.content.read()
        try:
            doc = json.loads(doc_content)
        except:
            try:
                multipart_content_type = request.headers['Content-Type']
                doc = {
                    part.headers[b'Content-Disposition']
                    .decode('utf-8')
                    .replace('form-data; name=', '')
                    .replace('"', '')
                    .split(';')[0]: part.content
                    for part in decoder.MultipartDecoder(
                        doc_content, multipart_content_type
                    ).parts
                }
            except Exception as e:
                doc = {}

        results = await Config.modules[module].methods[method](
            env=env, query=[request_args], doc=doc
        )

        logger.debug('Closing connection.')
        env['conn'].close()

        if 'return' not in results['args'] or results['args']['return'] == 'json':
            if 'return' in results['args']:
                del results['args']['return']
            headers['Content-Type'] = 'application/json; charset=utf-8'
            if results['status'] == 404:
                return aiohttp.web.Response(
                    status=results['status'],
                    headers=headers,
                    body=JSONEncoder()
                    .encode({'status': 404, 'msg': 'Requested content not found.'})
                    .encode('utf-8'),
                )
            return aiohttp.web.Response(
                status=results['status'],
                headers=headers,
                body=JSONEncoder().encode(results),
            )
        elif results['args']['return'] == 'file':
            del results['args']['return']
            expiry_time = datetime.datetime.utcnow() + datetime.timedelta(days=30)
            headers['lastModified'] = str(results['args']['docs'][0]['lastModified'])
            headers['Content-Type'] = results['args']['docs'][0]['type']
            headers['Cache-Control'] = 'public, max-age=31536000'
            headers['Expires'] = expiry_time.strftime('%a, %d %b %Y %H:%M:%S GMT')
            return aiohttp.web.Response(
                status=results['status'],
                headers=headers,
                body=results['args']['docs'][0]['content'],
            )
        elif results['args']['return'] == 'msg':
            del results['args']['return']
            headers['Content-Type'] = 'application/json; charset=utf-8'
            return aiohttp.web.Response(
                status=results['status'], headers=headers, body=results['msg']
            )

        headers['Content-Type'] = 'application/json; charset=utf-8'
        return aiohttp.web.Response(
            status=405,
            headers=headers,
            body=JSONEncoder().encode({'status': 405, 'msg': 'METHOD NOT ALLOWED'}),
        )

    async def websocket_handler(request: aiohttp.web.Request):
        conn = Data.create_conn()
        logger.info(
            f'Websocket connection starting with client at \'{request.remote}\''
        )
        ws = aiohttp.web.WebSocketResponse()
        await ws.prepare(request)

        env: NAWAH_ENV = {
            'id': len(sessions),
            'conn': conn,
            'REMOTE_ADDR': request.remote,
            'ws': ws,
            'watch_tasks': {},
            'init': False,
            'last_call': datetime.datetime.utcnow(),
            'quota': {
                'counter': Config.quota_anon_min,
                'last_check': datetime.datetime.utcnow(),
            },
        }
        sessions.append(env)
        try:
            env['HTTP_USER_AGENT'] = request.headers['user-agent']
            env['HTTP_ORIGIN'] = request.headers['origin']
        except:
            env['HTTP_USER_AGENT'] = ''
            env['HTTP_ORIGIN'] = ''

        logger.info(
            f'Websocket connection #\'{env["id"]}\' ready with client at \'{env["REMOTE_ADDR"]}\''
        )

        await ws.send_str(
            JSONEncoder().encode(
                {
                    'status': 200,
                    'msg': 'Connection ready',
                    'args': {'code': 'CORE_CONN_READY'},
                }
            )
        )

        async for msg in ws:
            if 'conn' not in env:
                await ws.close()
                break
            logger.info(
                f'Received new message from session #\'{env["id"]}\': {msg.data[:256]}'
            )
            if msg.type == aiohttp.WSMsgType.TEXT:
                logger.info(f'ip_quota on session #\'{env["id"]}\': {ip_quota}')
                logger.info(
                    f'session_quota: on session #\'{env["id"]}\': {env["quota"]}'
                )
                # [DOC] Check for IP quota
                if str(request.remote) not in ip_quota:
                    ip_quota[str(request.remote)] = {
                        'counter': Config.quota_ip_min,
                        'last_check': datetime.datetime.utcnow(),
                    }
                else:
                    if (
                        datetime.datetime.utcnow()
                        - ip_quota[str(request.remote)]['last_check']
                    ).seconds > 59:
                        ip_quota[str(request.remote)][
                            'last_check'
                        ] = datetime.datetime.utcnow()
                        ip_quota[str(request.remote)]['counter'] = Config.quota_ip_min
                    else:
                        if ip_quota[str(request.remote)]['counter'] - 1 <= 0:
                            logger.warning(
                                f'Denying Websocket request from \'{request.remote}\' for hitting IP quota.'
                            )
                            asyncio.create_task(
                                handle_msg(
                                    env=env,
                                    msg=msg,
                                    decline_quota='ip',
                                )
                            )
                            continue
                        else:
                            ip_quota[str(request.remote)]['counter'] -= 1
                # [DOC] Check for session quota
                if (
                    datetime.datetime.utcnow() - env['quota']['last_check']
                ).seconds > 59:
                    env['quota']['last_check'] = datetime.datetime.utcnow()
                    env['quota']['counter'] = (
                        (Config.quota_anon_min - 1)
                        if not env['session']
                        or env['session']['token'] == Config.anon_token
                        else (Config.quota_auth_min - 1)
                    )
                    asyncio.create_task(handle_msg(env=env, msg=msg))
                else:
                    if env['quota']['counter'] - 1 <= 0:
                        asyncio.create_task(
                            handle_msg(
                                env=env,
                                msg=msg,
                                decline_quota='session',
                            )
                        )
                        continue
                    else:
                        env['quota']['counter'] -= 1
                        asyncio.create_task(handle_msg(env=env, msg=msg))

        if 'id' in env.keys():
            await close_session(env['id'])

        return ws

    async def handle_msg(
        env: NAWAH_ENV,
        msg: aiohttp.WSMessage,
        decline_quota: str = None,
    ):
        try:
            env['last_call'] = datetime.datetime.utcnow()
            try:
                env['session']['token']
            except Exception:
                anon_user = _compile_anon_user()
                anon_session = _compile_anon_session()
                anon_session['user'] = anon_user
                env['session'] = anon_session
            res = json.loads(msg.data)
            try:
                res = jwt.decode(
                    res['token'], env['session']['token'], algorithms=['HS256']
                )
            except Exception:
                try:
                    res = jwt.decode(
                        res['token'], env['prev_session']['token'], algorithms=['HS256']
                    )
                except Exception:
                    await env['ws'].send_str(
                        JSONEncoder().encode(
                            {
                                'status': 403,
                                'msg': 'Request token is not accepted.',
                                'args': {
                                    'call_id': res['call_id']
                                    if 'call_id' in res.keys()
                                    else None,
                                    'code': 'CORE_REQ_INVALID_TOKEN',
                                },
                            }
                        )
                    )
                    if env['init'] == False:
                        await env['ws'].close()
                        return
                    else:
                        return

            # [DOC] Check if msg should be denied for quota hit
            if decline_quota == 'ip':
                await env['ws'].send_str(
                    JSONEncoder().encode(
                        {
                            'status': 429,
                            'msg': 'You have hit calls quota from this IP.',
                            'args': {
                                'call_id': res['call_id']
                                if 'call_id' in res.keys()
                                else None,
                                'code': 'CORE_REQ_IP_QUOTA_HIT',
                            },
                        }
                    )
                )
                return
            elif decline_quota == 'session':
                await env['ws'].send_str(
                    JSONEncoder().encode(
                        {
                            'status': 429,
                            'msg': 'You have hit calls quota.',
                            'args': {
                                'call_id': res['call_id']
                                if 'call_id' in res.keys()
                                else None,
                                'code': 'CORE_REQ_SESSION_QUOTA_HIT',
                            },
                        }
                    )
                )
                return

            logger.info(f'Decoded request: {JSONEncoder().encode(res)}')

            if 'endpoint' not in res.keys():
                await env['ws'].send_str(
                    JSONEncoder().encode(
                        {
                            'status': 400,
                            'msg': 'Request missing endpoint.',
                            'args': {
                                'call_id': res['call_id']
                                if 'call_id' in res.keys()
                                else None,
                                'code': 'CORE_REQ_NO_ENDPOINT',
                            },
                        }
                    )
                )
                return

            if env['init'] == False:
                if res['endpoint'] != 'conn/verify':
                    await env['ws'].send_str(
                        JSONEncoder().encode(
                            {
                                'status': 1008,
                                'msg': 'Request token is not accepted.',
                                'args': {
                                    'call_id': res['call_id']
                                    if 'call_id' in res.keys()
                                    else None,
                                    'code': 'CORE_REQ_NO_VERIFY',
                                },
                            }
                        )
                    )
                    await env['ws'].close()
                    return
                else:
                    # [TODO] Add condition to now allow a connection from a client with client_app = __sys
                    if len(Config.client_apps.keys()) and (
                        'doc' not in res.keys()
                        or 'app' not in res['doc'].keys()
                        or res['doc']['app'] not in Config.client_apps.keys()
                        or (
                            Config.client_apps[res['doc']['app']]['type'] == 'web'
                            and env['HTTP_ORIGIN']
                            not in Config.client_apps[res['doc']['app']]['origin']
                        )
                    ):
                        await env['ws'].send_str(
                            JSONEncoder().encode(
                                {
                                    'status': 1008,
                                    'msg': 'Request token is not accepted.',
                                    'args': {
                                        'call_id': res['call_id']
                                        if 'call_id' in res.keys()
                                        else None,
                                        'code': 'CORE_REQ_NO_VERIFY',
                                    },
                                }
                            )
                        )
                        await env['ws'].close()
                        return
                    else:
                        env['init'] = True
                        # [TODO] Add condition to now allow a connection from a client with client_app = __sys
                        if not Config.client_apps:
                            env['client_app'] = '__public'
                        else:
                            env['client_app'] = res['doc']['app']
                        logger.info(
                            f'Connection on session #\'{env["id"]}\' is verified.'
                        )
                        if Config.analytics_events['app_conn_verified']:
                            asyncio.create_task(
                                Config.modules['analytic'].create(
                                    skip_events=[Event.PERM],
                                    env=env,
                                    doc={
                                        'event': 'CONN_VERIFIED',
                                        'subevent': env['client_app'],
                                        'args': {
                                            'REMOTE_ADDR': env['REMOTE_ADDR'],
                                            'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
                                        },
                                    },
                                )
                            )
                        await env['ws'].send_str(
                            JSONEncoder().encode(
                                {
                                    'status': 200,
                                    'msg': 'Connection established',
                                    'args': {
                                        'call_id': res['call_id']
                                        if 'call_id' in res.keys()
                                        else None,
                                        'code': 'CORE_CONN_OK',
                                    },
                                }
                            )
                        )
                        return

            if res['endpoint'] == 'conn/close':
                logger.info(
                    f'Received connection close instructions on session #\'{env["id"]}\'.'
                )
                await env['ws'].close()
                return

            if res['endpoint'] == 'heart/beat':
                logger.info(
                    f'Received connection heartbeat on session #\'{env["id"]}\'.'
                )
                await env['ws'].send_str(
                    JSONEncoder().encode(
                        {
                            'status': 200,
                            'msg': 'Heartbeat received.',
                            'args': {
                                'call_id': res['call_id']
                                if 'call_id' in res.keys()
                                else None,
                                'code': 'CORE_HEARTBEAT_OK',
                            },
                        }
                    )
                )
                return

            res['endpoint'] = res['endpoint'].lower()

            if (
                res['endpoint'] in ['session/auth', 'session/reauth']
                and str(env['session']['_id']) != 'f00000000000000000000012'
            ):
                await env['ws'].send_str(
                    JSONEncoder().encode(
                        {
                            'status': 400,
                            'msg': 'You are already authed.',
                            'args': {
                                'call_id': res['call_id']
                                if 'call_id' in res.keys()
                                else None,
                                'code': 'CORE_SESSION_ALREADY_AUTHED',
                            },
                        }
                    )
                )
                return

            if (
                res['endpoint'] == 'session/signout'
                and str(env['session']['_id']) == 'f00000000000000000000012'
            ):
                await env['ws'].send_str(
                    JSONEncoder().encode(
                        {
                            'status': 400,
                            'msg': 'Singout is not allowed for \'__ANON\' user.',
                            'args': {
                                'call_id': res['call_id']
                                if 'call_id' in res.keys()
                                else None,
                                'code': 'CORE_SESSION_ANON_SIGNOUT',
                            },
                        }
                    )
                )
                return

            if 'query' not in res.keys():
                res['query'] = []
            if 'doc' not in res.keys():
                res['doc'] = {}
            if 'call_id' not in res.keys():
                res['call_id'] = ''

            request = {
                'call_id': res['call_id'],
                'sid': res['sid'] or False,
                'query': res['query'],
                'doc': res['doc'],
                'path': res['endpoint'].split('/'),
            }

            if len(request['path']) != 2:
                await env['ws'].send_str(
                    JSONEncoder().encode(
                        {
                            'status': 400,
                            'msg': 'Endpoint path is invalid.',
                            'args': {
                                'call_id': request['call_id'],
                                'code': 'CORE_REQ_INVALID_PATH',
                            },
                        }
                    )
                )
                return

            module = request['path'][0].lower()

            if module == 'watch' and request['path'][1].lower() == 'delete':
                logger.info(
                    'Received watch task delete request for: %s',
                    request['query'][0]['watch'],
                )
                try:
                    if request['query'][0]['watch'] == '__all':
                        for watch_task in env['watch_tasks'].values():
                            watch_task['stream'].close()
                            watch_task['task'].cancel()
                        await env['ws'].send_str(
                            JSONEncoder().encode(
                                {
                                    'status': 200,
                                    'msg': 'All watch tasks deleted.',
                                    'args': {
                                        'call_id': request['call_id'],
                                        'watch': list(env['watch_tasks'].keys()),
                                    },
                                }
                            )
                        )
                        env['watch_tasks'] = {}
                    else:
                        env['watch_tasks'][request['query'][0]['watch']][
                            'stream'
                        ].close()
                        env['watch_tasks'][request['query'][0]['watch']][
                            'task'
                        ].cancel()
                        await env['ws'].send_str(
                            JSONEncoder().encode(
                                {
                                    'status': 200,
                                    'msg': 'Watch task deleted.',
                                    'args': {
                                        'call_id': request['call_id'],
                                        'watch': [request['query'][0]['watch']],
                                    },
                                }
                            )
                        )
                        del env['watch_tasks'][request['query'][0]['watch']]
                except:
                    await env['ws'].send_str(
                        JSONEncoder().encode(
                            {
                                'status': 400,
                                'msg': 'Watch is invalid.',
                                'args': {
                                    'call_id': request['call_id'],
                                    'code': 'CORE_WATCH_INVALID_WATCH',
                                },
                            }
                        )
                    )
                return

            if module not in Config.modules.keys():
                await env['ws'].send_str(
                    JSONEncoder().encode(
                        {
                            'status': 400,
                            'msg': 'Endpoint module is invalid.',
                            'args': {
                                'call_id': request['call_id'],
                                'code': 'CORE_REQ_INVALID_MODULE',
                            },
                        }
                    )
                )
                return

            if request['path'][1].lower() not in Config.modules[module].methods.keys():
                await env['ws'].send_str(
                    JSONEncoder().encode(
                        {
                            'status': 400,
                            'msg': 'Endpoint method is invalid.',
                            'args': {
                                'call_id': request['call_id'],
                                'code': 'CORE_REQ_INVALID_METHOD',
                            },
                        }
                    )
                )
                return

            if Config.modules[module].methods[request['path'][1].lower()].get_method:
                await env['ws'].send_str(
                    JSONEncoder().encode(
                        {
                            'status': 400,
                            'msg': 'Endpoint method is a GET method.',
                            'args': {
                                'call_id': request['call_id'],
                                'code': 'CORE_REQ_GET_METHOD',
                            },
                        }
                    )
                )
                return

            if not request['sid']:
                request['sid'] = 'f00000000000000000000012'

            method = Config.modules[module].methods[request['path'][1].lower()]
            query = request['query']
            doc = request['doc']
            await _process_file_obj(doc=doc, modules=Config.modules, env=env)
            asyncio.create_task(
                method(
                    skip_events=[],
                    env=env,
                    query=query,
                    doc=doc,
                    call_id=request['call_id'],
                )
            )

        except Exception as e:
            logger.error(f'An error occurred. Details: {traceback.format_exc()}.')
            if Config.debug:
                await env['ws'].send_str(
                    JSONEncoder().encode(
                        {
                            'status': 500,
                            'msg': f'Unexpected error has occurred [{str(e)}].',
                            'args': {'code': 'CORE_SERVER_ERROR', 'err': str(e)},
                        }
                    )
                )
            else:
                await env['ws'].send_str(
                    JSONEncoder().encode(
                        {
                            'status': 500,
                            'msg': 'Unexpected error has occurred.',
                            'args': {'code': 'CORE_SERVER_ERROR'},
                        }
                    )
                )

    async def close_session(id):
        if sessions[id].keys():
            logger.info(
                f'Cleaning up watch tasks before connection for session #\'{sessions[id]["id"]}\' close.'
            )
            for watch_task in sessions[id]['watch_tasks'].values():
                try:
                    await watch_task['stream'].close()
                except Exception as e:
                    logger.error(f'stream close error: {e}')
                try:
                    watch_task['task'].cancel()
                except Exception as e:
                    logger.error(f'task close error: {e}')

            logger.info(
                f'Closing data connection for session #\'{sessions[id]["id"]}\''
            )
            sessions[id]['conn'].close()

            logger.info('Done closing data connection.')
            logger.info(f'Websocket connection status: {not sessions[id]["ws"].closed}')

            if not sessions[id]['ws'].closed:
                await sessions[id]['ws'].close()
            logger.info(f'Websocket connection for session #\'{id}\' closed.')

            sessions[id] = {}
        else:
            logger.info(f'Skipped closing session #\'{id}\'.')

    async def jobs_loop():
        while True:
            await asyncio.sleep(60)
            try:
                # [DOC] Connection Timeout Workflow
                logger.info('Time to check for sessions!')
                logger.info(f'Current sessions: {sessions}')
                for i in range(len(sessions)):
                    session = sessions[i]
                    if 'last_call' not in session.keys():
                        continue
                    if datetime.datetime.utcnow() > (
                        session['last_call']
                        + datetime.timedelta(seconds=Config.conn_timeout)
                    ):
                        logger.info(
                            f'Session #\'{session["id"]}\' with REMOTE_ADDR \'{session["REMOTE_ADDR"]}\' HTTP_USER_AGENT: \'{session["HTTP_USER_AGENT"]}\' is idle. Closing.'
                        )
                        await close_session(i)
            except Exception:
                logger.error(f'An error occurred. Details: {traceback.format_exc()}.')

            try:
                # [DOC] Calls Quota Workflow - Clean-up Sequence
                logger.info('Time to check for IPs quotas!')
                del_ip_quota = []
                for ip in ip_quota.keys():
                    if (
                        datetime.datetime.utcnow() - ip_quota[ip]['last_check']
                    ).seconds > 59:
                        logger.info(
                            f'IP \'{ip}\' with quota \'{ip_quota[ip]["counter"]}\' is idle. Cleaning-up.'
                        )
                        del_ip_quota.append(ip)
                for ip in del_ip_quota:
                    del ip_quota[ip]
            except Exception:
                logger.error(f'An error occurred. Details: {traceback.format_exc()}.')

            try:
                # [DOC] Jobs Workflow
                current_time = datetime.datetime.utcnow().isoformat()[:16]
                logger.info('Time to check for jobs!')
                for job_name in Config.jobs:
                    job = Config.jobs[job_name]
                    logger.info(f'Checking: {job_name}')
                    if job._disabled:
                        logger.info('-Job is disabled. Skipping..')
                        continue
                    # [DOC] Check if job is scheduled for current_time
                    if current_time == job._next_time:
                        logger.info('-Job is due, running!')
                        # [DOC] Update job next_time
                        job._next_time = datetime.datetime.fromtimestamp(
                            job._cron_schedule.get_next(), datetime.timezone.utc
                        ).isoformat()[:16]
                        try:
                            job.job(env=Config._sys_env)
                        except Exception as e:
                            logger.error(
                                'Job \'{job_name}\' has failed with exception.'
                            )
                            logger.error('Exception details:')
                            logger.error(traceback.format_exc())
                            if job.prevent_disable:
                                logger.warning(
                                    '-Detected job prevent_disable. Skipping disabling job..'
                                )
                            else:
                                logger.warning('-Disabling job.')
                                job._disabled = True
                    else:
                        logger.info('-Not yet due.')
            except Exception:
                logger.error(f'An error occurred. Details: {traceback.format_exc()}.')

            try:
                logger.info('Time to check for files timeout!')
                files_results = await Config.modules['file'].delete(
                    skip_events=[Event.PERM],
                    env=Config._sys_env,
                    query=[
                        {
                            'create_time': {
                                '$lt': (
                                    datetime.datetime.utcnow()
                                    - datetime.timedelta(
                                        seconds=Config.file_upload_timeout
                                    )
                                ).isoformat()
                            }
                        }
                    ],
                )
                logger.info('Files timeout results:')
                logger.info(f'-status: {files_results["status"]}')
                logger.info(f'-msg: {files_results["msg"]}')
                logger.info(f'-args.docs: {files_results["args"]["docs"]}')
            except Exception:
                logger.error(f'An error occurred. Details: {traceback.format_exc()}.')

    def create_error_middleware(overrides):
        @aiohttp.web.middleware
        async def error_middleware(request, handler):
            try:
                response = await handler(request)
                override = overrides.get(response.status)
                if override:
                    return await override(request)
                return response
            except aiohttp.web.HTTPException as ex:
                override = overrides.get(ex.status)
                if override:
                    return await override(request)
                raise

        return error_middleware

    async def web_loop():
        app = aiohttp.web.Application()
        app.middlewares.append(
            create_error_middleware(
                {
                    404: not_found_handler,
                    405: not_allowed_handler,
                }
            )
        )
        app.router.add_route('GET', '/', root_handler)
        app.router.add_route('*', '/ws', websocket_handler)
        for route in get_routes:
            app.router.add_route('GET', route, http_handler)
        for route in post_routes:
            app.router.add_route('POST', route, http_handler)
            app.router.add_route('OPTIONS', route, http_handler)
        logger.info('Welcome to Nawah')
        await aiohttp.web.run_app(app, host='0.0.0.0', port=Config.port)

    async def loop_gather():
        await asyncio.gather(jobs_loop(), web_loop())

    try:
        asyncio.run(loop_gather())
    except KeyboardInterrupt:
        if time.localtime().tm_hour >= 21 or time.localtime().tm_hour <= 4:
            msg = 'night'
        elif time.localtime().tm_hour >= 18:
            msg = 'evening'
        elif time.localtime().tm_hour >= 12:
            msg = 'afternoon'
        elif time.localtime().tm_hour >= 5:
            msg = 'morning'
        logger.info(f'Have a great {msg}!')
        exit()



from typing import Any, Dict

from nawah.classes import ATTR, NAWAH_DOC, NAWAH_ENV
from nawah.config import Config


async def create(
    *,
    env: NAWAH_ENV,
    collection_name: str,
    attrs: Dict[str, ATTR],
    doc: NAWAH_DOC,
) -> Dict[str, Any]:
    collection = env['conn'][Config.data_name][collection_name]
    results = await collection.insert_one(doc)
    _id = results.inserted_id
    return {'count': 1, 'docs': [{'_id': _id}]}



import datetime
import re
from typing import Any, Dict

import mock
import pytest
from bson import ObjectId

from nawah import base_module, registry, testing
from nawah.base_method._check_permissions import _check_permissions
from nawah.base_method._validate_args import _validate_args
from nawah.classes import (ATTR, InvalidCallArgsException,
                           InvalidPermissionsExcpetion, MethodException, Query)
from nawah.config import Config
from nawah.enums import NAWAH_VALUES, Event

if testing.NAWAH_TESTING:
    import logging

    logger = logging.getLogger('nawah')
    logger.error('Running both unit and integration tests in one run is not supported.')
    exit(1)  # Running both unit and integration tests in one run is not supported.

testing.NAWAH_TESTING = 'package'


class MockBaseModule:
    _mock_nawah_module = True

    collection = None
    attrs = None
    diff = None
    create_draft = None
    update_draft = None
    defaults = None
    unique_attrs = None
    extns = None
    privileges = None
    methods = None
    cache = None
    analytics = None

    package_name = None
    module_name = None

    def _pre_initialise(self):
        raise NotImplementedError()

    def _initialise(self):
        raise NotImplementedError()

    def status(self, *, status, msg, args=None):
        status_dict = {'status': status, 'msg': msg, 'args': {}}
        if args:
            if 'code' in args.keys():
                args[
                    'code'
                ] = f'{self.package_name.upper()}_{self.module_name.upper()}_{args["code"]}'
            status_dict['args'] = args
        return status_dict

    def exception(self, *, status, msg, args=None):
        status_dict = {'status': status, 'msg': msg, 'args': {}}
        if args:
            if 'code' in args.keys():
                args[
                    'code'
                ] = f'{self.package_name.upper()}_{self.module_name.upper()}_{args["code"]}'
            status_dict['args'] = args
        return MethodException(status_dict)

    async def pre_read(self, skip_events, env, query, doc, payload):
        raise NotImplementedError('Handler pre_read is not implemented')

    async def on_read(self, results, skip_events, env, query, doc, payload):
        raise NotImplementedError('Handler on_read is not implemented')

    async def read(self, skip_events, env, query, doc):
        raise NotImplementedError('Method read is not implemented')

    async def pre_watch(self, skip_events, env, query, doc, payload):
        raise NotImplementedError('Handler pre_watch is not implemented')

    async def on_watch(self, results, skip_events, env, query, doc, payload):
        raise NotImplementedError('Handler on_watch is not implemented')

    async def watch(self, skip_events, env, query, doc):
        raise NotImplementedError('Method watch is not implemented')

    async def pre_create(self, skip_events, env, query, doc, payload):
        raise NotImplementedError('Handler pre_create is not implemented')

    async def on_create(self, results, skip_events, env, query, doc, payload):
        raise NotImplementedError('Handler on_create is not implemented')

    async def create(self, skip_events, env, query, doc):
        raise NotImplementedError('Method create is not implemented')

    async def pre_update(self, skip_events, env, query, doc, payload):
        raise NotImplementedError('Handler pre_update is not implemented')

    async def on_update(self, results, skip_events, env, query, doc, payload):
        raise NotImplementedError('Handler on_update is not implemented')

    async def update(self, skip_events, env, query, doc):
        raise NotImplementedError('Method update is not implemented')

    async def pre_delete(self, skip_events, env, query, doc, payload):
        raise NotImplementedError('Handler pre_delete is not implemented')

    async def on_delete(self, results, skip_events, env, query, doc, payload):
        raise NotImplementedError('Handler on_delete is not implemented')

    async def delete(self, skip_events, env, query, doc):
        raise NotImplementedError('Method delete is not implemented')

    async def pre_create_file(self, skip_events, env, query, doc, payload):
        raise NotImplementedError('Handler pre_create_file is not implemented')

    async def on_create_file(self, results, skip_events, env, query, doc, payload):
        raise NotImplementedError('Handler on_create_file is not implemented')

    async def create_file(self, skip_events, env, query, doc):
        raise NotImplementedError('Method create_file is not implemented')

    async def pre_delete_file(self, skip_events, env, query, doc, payload):
        raise NotImplementedError('Handler pre_delete_file is not implemented')

    async def on_delete_file(self, results, skip_events, env, query, doc, payload):
        raise NotImplementedError('Handler on_delete_file is not implemented')

    async def delete_file(self, skip_events, env, query, doc):
        raise NotImplementedError('Method delete_file is not implemented')

    async def pre_retrieve_file(self, skip_events, env, query, doc, payload):
        raise NotImplementedError('Handler pre_retrieve_file is not implemented')

    async def on_retrieve_file(self, results, skip_events, env, query, doc, payload):
        raise NotImplementedError('Handler on_retrieve_file is not implemented')

    async def retrieve_file(self, skip_events, env, query, doc):
        raise NotImplementedError('Method retrieve_file is not implemented')


class MockRegistry(registry.Registry):
    modules: Dict[str, MockBaseModule] = {}

    @staticmethod
    def module(module):
        try:
            return MockRegistry.modules[module]
        except KeyError:
            raise registry.InvalidModuleException(module=module)


base_module.BaseModule = MockBaseModule  # type: ignore
registry.Registry = MockRegistry  # type: ignore


def pytest_configure(config):
    config.addinivalue_line(
        'markers',
        'setup_test(modules, l10n, vars, types): Config mocked environment for testing.',
    )


def pytest_runtest_setup(item):
    for marker in item.iter_markers(name='setup_test'):
        setup_test(**marker.kwargs)


def setup_test(*, modules=None, l10n_dicts=None, vars=None, types=None):
    MockRegistry.modules = {}
    Config.l10n = {}
    Config.vars = {}
    Config.types = {}

    if modules:
        for module_class in modules.keys():
            module = module_class()
            module.package_name = (
                module.__module__.replace('modules.', '').upper().split('.')[-2]
            )
            module.module_name = re.sub(
                r'([A-Z])',
                r'_\1',
                module.__class__.__name__[0].lower() + module.__class__.__name__[1:],
            ).lower()
            for method_name in modules[module_class].keys():
                mock_method = mock.AsyncMock()
                mock_method.return_value = modules[module_class][method_name]
                setattr(module, method_name, mock_method)

            MockRegistry.modules[module.module_name] = module

    if l10n_dicts:
        Config.l10n = l10n_dicts

    if vars:
        Config.vars = vars

    if types:
        Config.types = types


@pytest.fixture
def mock_env():
    def _(*, session=None, user=None, privileges=None):
        if not session:
            session = {
                '_id': ObjectId('f00000000000000000000012'),
                'user': None,
                'host_add': '127.0.0.1',
                'user_agent': '__ANON_TOKEN_000000000000000000000000',
                'timestamp': '1970-01-01T00:00:00',
                'expiry': '1970-01-01T00:00:00',
                'token': '__ANON_TOKEN_000000000000000000000000',
                'token_hash': '__ANON_TOKEN_000000000000000000000000',
            }

        if not user:
            user = {
                '_id': ObjectId('f00000000000000000000011'),
                'name': {'na_NA': '__ANON'},
                'groups': [],
                'privileges': None,
                'locale': 'na_NA',
            }

        if not privileges:
            privileges = {}

        session['user'] = user
        user['privileges'] = privileges

        return {
            'id': 0,
            'conn': None,
            'REMOTE_ADDR': '127.0.0.1',
            'ws': None,
            'watch_tasks': {},
            'init': False,
            'last_call': datetime.datetime.utcnow(),
            'quota': {
                'counter': 0,
                'last_check': datetime.datetime.utcnow(),
            },
            'HTTP_USER_AGENT': '',
            'HTTP_ORIGIN': '',
            'session': session,
        }

    return _


@pytest.fixture
def call_handler_pre(mock_env):
    async def _(
        *, module, handler, skip_events=[], env=None, query=[], doc={}, payload={}
    ):
        if not env:
            env = mock_env()
        query = Query(query)
        return await getattr(MockRegistry.modules[module], handler)(
            skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
        )

    return _


@pytest.fixture
def call_handler_on(mock_env):
    async def _(
        *,
        module,
        handler,
        results,
        skip_events=[],
        env=None,
        query=[],
        doc={},
        payload={},
    ):
        if not env:
            env = mock_env()
        query = Query(query)

        return await getattr(MockRegistry.modules[module], handler)(
            results=results,
            skip_events=skip_events,
            env=env,
            query=query,
            doc=doc,
            payload=payload,
        )

    return _


@pytest.fixture
def call_method_check_permissions(mock_env):
    async def _(*, module, method, skip_events=[], env=None, query=[], doc={}):
        if not env:
            env = mock_env()
        query = Query(query)

        permissions = await _check_permissions(
            skip_events=skip_events,
            env=env,
            query=query,
            doc=doc,
            module=MockRegistry.modules[module],
            permissions=MockRegistry.modules[module].methods[method].permissions,
        )

        return permissions

    return _


@pytest.fixture
def call_method_validate_args():
    async def _(*, args, args_list_label, args_list):
        await _validate_args(
            args=args,
            args_list_label=args_list_label,
            args_list=args_list,
        )

    return _


@pytest.fixture
def call_method(mock_env, call_method_check_permissions, call_method_validate_args):
    async def _(*, module, method, skip_events=[], env=None, query=[], doc={}):
        if not env:
            env = mock_env()
        query = Query(query)

        # [DOC] Following code is actual copy of BaseModule.__call__ method as to replicate the complete process of calling an initialised Nawah module method

        if Event.PERM not in skip_events:
            permissions_check = await call_method_check_permissions(
                module=module,
                method=method,
                skip_events=skip_events,
                env=env,
                query=query,
                doc=doc,
            )

            if type(permissions_check['query_mod']) == dict:
                permissions_check['query_mod'] = [permissions_check['query_mod']]
            for i in range(len(permissions_check['query_mod'])):
                if type(permissions_check['query_mod'][i]) == dict:
                    query_set_list = [permissions_check['query_mod'][i]]
                elif type(permissions_check['query_mod'][i]) == list:
                    query_set_list = permissions_check['query_mod'][i]
                for query_set in query_set_list:
                    del_args = []
                    for attr in query_set.keys():
                        if query_set[attr] == None:
                            del_args.append(attr)
                    for attr in del_args:
                        del query_set[attr]
            query.append(permissions_check['query_mod'])

            del_args = []
            for attr in permissions_check['doc_mod'].keys():
                if permissions_check['doc_mod'][attr] == None:
                    permissions_check['doc_mod'][attr] = NAWAH_VALUES.NONE_VALUE
            for attr in del_args:
                del permissions_check['doc_mod'][attr]
            doc.update(permissions_check['doc_mod'])
            doc = {
                attr: doc[attr]
                for attr in doc.keys()
                if doc[attr] != NAWAH_VALUES.NONE_VALUE
            }

        if Event.ARGS not in skip_events:
            if args_list := MockRegistry.modules[module].methods[method].query_args:
                if type(args_list) != list:
                    args_list = [args_list]

                # [DOC] Call ATTR.validate_type on query_args to initialise Attr Type TYPE
                for args_set in args_list:
                    for attr in args_set.keys():
                        ATTR.validate_type(attr_type=args_set[attr])

                await _validate_args(
                    args=query,
                    args_list_label='query',
                    args_list=args_list,
                )

            if args_list := MockRegistry.modules[module].methods[method].doc_args:
                if type(args_list) != list:
                    args_list = [args_list]

                # [DOC] Call ATTR.validate_type on query_args to initialise Attr Type TYPE
                for args_set in args_list:
                    for attr in args_set.keys():
                        ATTR.validate_type(attr_type=args_set[attr])

                await _validate_args(
                    args=doc,
                    args_list_label='doc',
                    args_list=args_list,
                )

        for arg in doc.keys():
            if type(doc[arg]) == dict and '_id' in doc[arg]:
                doc[arg] = doc[arg]['_id']  # type: ignore

        # [DOC] check if $soft oper is set to add it to events
        if '$soft' in query and query['$soft'] == True:
            skip_events.append(Event.SOFT)
            del query['$soft']

        # [DOC] check if $extn oper is set to add it to events
        if '$extn' in query and query['$extn'] == False:
            skip_events.append(Event.EXTN)
            del query['$extn']

        return await getattr(MockRegistry.modules[module], method)(
            skip_events=skip_events, env=env, query=query, doc=doc
        )

    return _



from nawah import testing
from nawah.config import Config
from nawah.data import create_conn
from nawah.cli import launch
from nawah.utils import _import_modules, _config_data, generate_attr

import pytest, argparse

if testing.NAWAH_TESTING:
    import logging

    logger = logging.getLogger('nawah')
    logger.error('Running both unit and integration tests in one run is not supported.')
    exit(1)  # Running both unit and integration tests in one run is not supported.

testing.NAWAH_TESTING = 'app'

test_env = None


@pytest.fixture
def setup_test():
    async def _():
        global test_env

        if test_env:
            raise Exception('Fixture setup_test can only be run once.')

        Config.test = True
        launch(args=argparse.Namespace(env=None, debug=True), custom_launch='test')

        await _import_modules()
        await _config_data()

        test_env = Config._sys_env

    return _


@pytest.fixture
def env():
    def _():
        global test_env

        return test_env

    return _



'''Exposes package members to other packages'''

from ._attr import ATTR, ATTRS_TYPES_ARGS, SPECIAL_ATTRS
from ._exceptions import *
from ._json_encoder import JSONEncoder
from ._module import (ANALYTIC, CACHE, CACHE_CONDITION, EXTN, METHOD,
                      ON_HANDLER_RETURN, PERM, PRE_HANDLER_RETURN, RESULTS,
                      RESULTS_ARGS)
from ._package import (ANALYTICS_EVENTS, APP_CONFIG, CLIENT_APP, JOB, L10N,
                       PACKAGE_CONFIG, SYS_DOC, USER_SETTING)
from ._query import Query
from ._types import (IP_QUOTA, NAWAH_DOC, NAWAH_ENV, NAWAH_EVENTS, NAWAH_QUERY,
                     NAWAH_QUERY_SPECIAL, NAWAH_QUERY_SPECIAL_GROUP,
                     WATCH_TASK)



import datetime
import logging
import os
import redis
import time
from typing import Any, Dict, Union, cast

import requests
from bson import ObjectId
from croniter import croniter
from passlib.hash import pbkdf2_sha512

from nawah.classes import (APP_CONFIG, NAWAH_DOC, PACKAGE_CONFIG, SYS_DOC,
                           USER_SETTING)
from nawah.config import Config
from nawah.enums import Event

from ._attr import _deep_update

logger = logging.getLogger('nawah')


def _process_config(*, config: Union[APP_CONFIG, PACKAGE_CONFIG], pkgname: str = None):
    if type(config) not in [APP_CONFIG, PACKAGE_CONFIG]:
        logger.error(f'Config object of type \'{type(config)}\' is invalid. Exiting.')
        exit(1)

    if type(config) == PACKAGE_CONFIG and not pkgname:
        logger.error(
            'Provided Config object of type \'PACKAGE_CONFIG\' without \'pkgname\'. Exiting.'
        )
        exit(1)

    app_only_attrs = APP_CONFIG.__annotations__.keys()

    for config_attr in dir(config):
        config_attr_val = getattr(config, config_attr)

        # [DOC] Check existence of of api_level, version Config Attrs
        if type(config) == PACKAGE_CONFIG and config_attr in ['api_level', 'version']:
            if config_attr_val == None:
                logger.error(
                    f'Package \'{pkgname}\' is missing \'{config_attr}\' Config Attr. Exiting.'
                )
                exit(1)
            # [DOC] Check type of api_level, version Config Attrs
            elif type(config_attr_val) != str:
                logger.error(
                    f'Package \'{pkgname}\' is having invalid type of \'{config_attr}\'. Exiting.'
                )
                exit(1)
            else:
                # [DOC] Update corresponding Config
                getattr(Config, f'packages_{config_attr}s')[pkgname] = config_attr_val
        # [DOC] Skip non Config Attr attrs
        elif (
            config_attr.startswith('__')
            or config_attr_val == None
            or config_attr in app_only_attrs
        ):
            continue
        # [DOC] For vars_types Config Attr, preserve package name for debugging purposes
        elif config_attr == 'vars_types':
            for var in config_attr_val.keys():
                Config.vars_types[var] = {
                    'package': pkgname,
                    'type': config_attr_val[var],
                }
        elif type(config_attr_val) == list:
            for j in config_attr_val:
                getattr(Config, config_attr).append(j)
            if config_attr == 'locales':
                Config.locales = list(set(Config.locales))
        elif type(config_attr_val) == dict:
            if not getattr(Config, config_attr):
                setattr(Config, config_attr, {})
            _deep_update(
                target=getattr(Config, config_attr), new_values=config_attr_val
            )
        else:
            setattr(Config, config_attr, config_attr_val)


async def _config_data():
    from nawah.utils import generate_attr

    # [TODO] Add validator for user_attrs, user_doc_settings
    # [DOC] Check app packages
    if Config._app_packages or len(Config.packages_versions.keys()) > 2:
        logger.debug(
            'Found \'_app_packages\' Config Attr. Attempting to validate all loaded packages are matching _app_packages Config Attr value.'
        )

        Config._app_packages['core'] = Config.packages_versions['core']
        Config._app_packages[Config._app_default_package] = Config.packages_versions[
            Config._app_default_package
        ]

        missing_packages = [
            package
            for package in Config._app_packages.keys()
            if package not in Config.packages_versions.keys()
        ]
        if missing_packages:
            logger.error(
                f'At least one package is missing that is required by app. Missing package[s]: \'{", ".join(missing_packages)}\'. Exiting.'
            )
            exit(1)

        extra_packages = [
            package
            for package in Config.packages_versions.keys()
            if package not in Config._app_packages.keys()
        ]
        if extra_packages:
            logger.error(
                f'At least one extra package is present in \'packages\' folder that is not required by app. Extra package[s]: \'{", ".join(extra_packages)}\'. Exiting.'
            )
            exit(1)

        # [DOC] Check for version mismatch
        for package, version in Config._app_packages.items():
            # [DOC] Skip core and default_packages
            if package in ['core', Config._app_default_package]:
                continue
            if version != Config.packages_versions[package]:
                logger.error(
                    f'Package \'{package}\' version \'{Config.packages_versions[package]}\' is added to app but not matching required version \'{version}\'. Exiting.'
                )
                exit(1)

    # [DOC] Check API version
    if not Config.packages_api_levels:
        logger.warning(
            'No API-level specified for the app. Nawah would continue to run the app, but the developer should consider adding API-level to eliminate specs mismatch.'
        )
    else:
        nawah_level = '.'.join(Config._nawah_version.split('.')[0:2])
        for package, api_level in Config.packages_api_levels.items():
            if api_level != nawah_level:
                logger.error(
                    f'Nawah framework is on API-level \'{nawah_level}\', but the app package \'{package}\' requires API-level \'{api_level}\'. Exiting.'
                )
                exit(1)
        try:
            versions = (
                (
                    requests.get(
                        'https://raw.githubusercontent.com/masaar/nawah_versions/master/versions.txt'
                    ).content
                )
                .decode('utf-8')
                .split('\n')
            )
            version_detected = ''
            for version in versions:
                if version.startswith(f'{nawah_level}.'):
                    if version_detected and int(version.split('.')[-1]) < int(
                        version_detected.split('.')[-1]
                    ):
                        continue
                    version_detected = version
            if version_detected and version_detected != Config._nawah_version:
                logger.warning(
                    f'Your app is using Nawah version \'{Config._nawah_version}\' while newer version \'{version_detected}\' of the API-level is available. Please, update.'
                )
        except:
            logger.warning(
                'An error occurred while attempting to check for latest update to Nawah. Please, check for updates on your own.'
            )

    # [DOC] Check for jobs
    if Config.jobs:
        # [DOC] Check jobs schedule validity
        Config._jobs_base = datetime.datetime.utcnow()
        for job_name in Config.jobs.keys():
            job = Config.jobs[job_name]
            if not croniter.is_valid(job.schedule):
                logger.error(
                    f'Job with schedule \'{job_name}\' schedule is invalid. Exiting.'
                )
                exit(1)

            job._cron_schedule = croniter(job.schedule, Config._jobs_base)
            job._next_time = datetime.datetime.fromtimestamp(
                job._cron_schedule.get_next(), datetime.timezone.utc
            ).isoformat()[:16]

    # [DOC] Check for presence of user_auth_attrs
    if not Config.user_attrs.keys():
        logger.error('No \'user_attrs\' are provided. Exiting.')
        exit(1)

    # [DOC] Check default values
    security_warning = '[SECURITY WARNING] {config_attr} is not explicitly set. It has been defaulted to \'{val}\' but in production environment you should consider setting it to your own to protect your app from breaches.'
    if Config.admin_password == '__ADMIN':
        logger.warning(
            security_warning.format(config_attr='Admin password', val='__ADMIN')
        )
    if Config.anon_token == '__ANON_TOKEN_f00000000000000000000012':
        logger.warning(
            security_warning.format(
                config_attr='Anon token',
                val='__ANON_TOKEN_f00000000000000000000012',
            )
        )

    # [DOC] Check for Env Vars
    attrs_defaults = {
        'data_server': 'mongodb://localhost',
        'data_name': 'nawah_data',
        'data_ssl': False,
        'data_ca_name': False,
        'data_ca': False,
        'emulate_test': False,
    }
    for attr_name in attrs_defaults.keys():
        attr_val = getattr(Config, attr_name)
        if type(attr_val) == str and attr_val.startswith('$__env.'):
            logger.debug(f'Detected Env Variable for config attr \'{attr_name}\'')
            if not os.getenv(attr_val[7:]):
                logger.warning(
                    f'Couldn\'t read Env Variable for config attr \'{attr_name}\'. Defaulting to \'{attrs_defaults[attr_name]}\''
                )
                setattr(Config, attr_name, attrs_defaults[attr_name])
            else:
                # [DOC] Set data_ssl to True rather than string Env Variable value
                if attr_name == 'ssl':
                    attr_val = True
                else:
                    attr_val = os.getenv(attr_val[7:])
                logger.warning(
                    f'Setting Env Variable for config attr \'{attr_name}\' to \'{attr_val}\''
                )
                setattr(Config, attr_name, attr_val)

    # [DOC] Check SSL settings
    if Config.data_ca and Config.data_ca_name:
        __location__ = os.path.realpath(os.path.join('.'))
        if not os.path.exists(os.path.join(__location__, 'certs')):
            os.makedirs(os.path.join(__location__, 'certs'))
        with open(os.path.join(__location__, 'certs', Config.data_ca_name), 'w') as f:
            f.write(Config.data_ca)

    from nawah import data as Data

    # [DOC] Create default env dict
    anon_user = _compile_anon_user()
    anon_session = _compile_anon_session()
    anon_session['user'] = anon_user
    Config._sys_conn = Data.create_conn()
    Config._sys_env = {
        'conn': Config._sys_conn,
        'REMOTE_ADDR': '127.0.0.1',
        'HTTP_USER_AGENT': 'Nawah',
        'client_app': '__sys',
        'session': anon_session,
        'watch_tasks': {},
    }

    if Config.data_azure_mongo:
        for module in Config.modules.keys():
            try:
                if Config.modules[module].collection:
                    logger.debug(
                        f'Attempting to create shard collection: {Config.modules[module].collection}.'
                    )
                    Config._sys_conn[Config.data_name].command(
                        'shardCollection',
                        f'{Config.data_name}.{Config.modules[module].collection}',
                        key={'_id': 'hashed'},
                    )
                else:
                    logger.debug(f'Skipping service module: {module}.')
            except Exception as err:
                logger.error(err)

    if Config.cache_server:
        Config._sys_cache = redis.Redis(Config.cache_server, db=Config.cache_db).json()

    # [DOC] Check test mode
    if Config.test:
        logger.debug('Test mode detected.')
        logger.setLevel(logging.DEBUG)
        __location__ = os.path.realpath(os.path.join('.'))
        if not os.path.exists(os.path.join(__location__, 'tests')):
            os.makedirs(os.path.join(__location__, 'tests'))
        for module in Config.modules.keys():
            module_collection = Config.modules[module].collection
            if module_collection:
                logger.debug(
                    f'Updating collection name \'{module_collection}\' of module {module}'
                )
                module_collection = Config.modules[
                    module
                ].collection = f'test_{module_collection}'
                if Config.test:
                    logger.debug(f'Flushing test collection \'{module_collection}\'')
                    await Data.drop(
                        env=Config._sys_env,
                        collection_name=module_collection,
                    )
            else:
                logger.debug(f'Skipping service module {module}')

    # [DOC] Test user_settings
    logger.debug('Testing user_settings.')
    if Config.user_settings:
        for user_setting in Config.user_settings.keys():
            logger.debug(f'Testing {user_setting}')
            if type(Config.user_settings[user_setting]) != USER_SETTING:
                logger.error(
                    f'Invalid Config Attr \'user_settings\' with key \'{user_setting}\' of type \'{type(Config.user_settings[user_setting])}\' with required type \'USER_SETTING\'. Exiting.'
                )
                exit(1)

            # [DOC] Validate USER_SETTING
            Config.user_settings[user_setting]._validate()

    # [DOC] Checking users collection
    # [TODO] Updated sequence to handle users
    logger.debug('Testing users collection.')
    user_results = await Config.modules['user'].read(
        skip_events=[Event.PERM, Event.ON],
        env=Config._sys_env,
        query=[{'_id': 'f00000000000000000000010'}],
    )
    if not user_results['args']['count']:
        logger.debug('ADMIN user not found, creating it.')
        # [DOC] Prepare base ADMIN user doc
        admin_create_doc = {
            '_id': ObjectId('f00000000000000000000010'),
            'name': {Config.locale: '__ADMIN'},
            'groups': [],
            'privileges': {'*': ['*']},
            'locale': Config.locale,
        }
        # [DOC] Update ADMIN user doc with admin_doc Config Attr
        admin_create_doc.update(Config.admin_doc)

        for auth_attr in Config.user_attrs.keys():
            admin_create_doc[f'{auth_attr}_hash'] = pbkdf2_sha512.using(
                rounds=100000
            ).hash(
                f'{auth_attr}{admin_create_doc[auth_attr]}{Config.admin_password}{Config.anon_token}'.encode(
                    'utf-8'
                )
            )
        admin_results = await Config.modules['user'].create(
            skip_events=[Event.PERM],
            env=Config._sys_env,
            doc=admin_create_doc,
        )
        logger.debug(f'ADMIN user creation results: {admin_results}')
        if admin_results['status'] != 200:
            logger.error('Config step failed. Exiting.')
            exit(1)
    elif not Config.force_admin_check:
        logger.warning(
            'ADMIN user found, skipping check due to force_admin_check Config Attr.'
        )
    else:
        logger.warning(
            'ADMIN user found, checking it due to force_admin_check Config Attr.'
        )
        admin_doc = user_results['args']['docs'][0]
        admin_doc_update = {}
        for attr in Config.admin_doc.keys():
            if (
                attr not in admin_doc
                or not admin_doc[attr]
                or Config.admin_doc[attr] != admin_doc[attr]
            ):
                if (
                    type(Config.admin_doc[attr]) == dict
                    and Config.locale in Config.admin_doc[attr].keys()
                    and type(admin_doc[attr]) == dict
                    and (
                        (
                            Config.locale in admin_doc[attr].keys()
                            and Config.admin_doc[attr][Config.locale]
                            == admin_doc[attr][Config.locale]
                        )
                        or (Config.locale not in admin_doc[attr].keys())
                    )
                ):
                    continue
                logger.debug(f'Detected change in \'admin_doc.{attr}\' Config Attr.')
                admin_doc_update[attr] = Config.admin_doc[attr]
        for auth_attr in Config.user_attrs.keys():
            auth_attr_hash = pbkdf2_sha512.using(rounds=100000).hash(
                f'{auth_attr}{admin_doc[auth_attr]}{Config.admin_password}{Config.anon_token}'.encode(
                    'utf-8'
                )
            )
            if (
                f'{auth_attr}_hash' not in admin_doc
                or auth_attr_hash != admin_doc[f'{auth_attr}_hash']
            ):
                logger.debug(f'Detected change in \'admin_password\' Config Attr.')
                admin_doc_update[f'{auth_attr}_hash'] = auth_attr_hash
        if len(admin_doc_update.keys()):
            logger.debug(
                f'Attempting to update ADMIN user with doc: \'{admin_doc_update}\''
            )
            admin_results = await Config.modules['user'].update(
                skip_events=[Event.PERM, Event.PRE, Event.ON],
                env=Config._sys_env,
                query=[{'_id': ObjectId('f00000000000000000000010')}],
                doc=admin_doc_update,
            )
            logger.debug(f'ADMIN user update results: {admin_results}')
            if admin_results['status'] != 200:
                logger.error('Config step failed. Exiting.')
                exit(1)
        else:
            logger.debug('ADMIN user is up-to-date.')

    Config._sys_docs[ObjectId('f00000000000000000000010')] = SYS_DOC(module='user')

    # [DOC] Test if ANON user exists
    user_results = await Config.modules['user'].read(
        skip_events=[Event.PERM, Event.ON],
        env=Config._sys_env,
        query=[{'_id': 'f00000000000000000000011'}],
    )
    if not user_results['args']['count']:
        logger.debug('ANON user not found, creating it.')
        anon_results = await Config.modules['user'].create(
            skip_events=[Event.PERM, Event.PRE, Event.ON],
            env=Config._sys_env,
            doc=_compile_anon_user(),
        )
        logger.debug(f'ANON user creation results: {anon_results}')
        if anon_results['status'] != 200:
            logger.error('Config step failed. Exiting.')
            exit(1)
    else:
        logger.debug('ANON user found, checking it.')
        anon_doc = _compile_anon_user()
        anon_doc_update = {}
        for attr in Config.user_attrs.keys():
            if attr not in anon_doc or not anon_doc[attr]:
                logger.debug(f'Detected change in \'anon_doc.{attr}\' Config Attr.')
                anon_doc_update[attr] = generate_attr(attr_type=Config.user_attrs[attr])
        for module in Config.anon_privileges.keys():
            if module not in anon_doc or set(anon_doc[module]) != set(
                Config.anon_privileges[module]
            ):
                logger.debug(f'Detected change in \'anon_privileges\' Config Attr.')
                anon_doc_update[f'privileges.{module}'] = Config.anon_privileges[module]
        for auth_attr in Config.user_attrs.keys():
            if (
                f'{auth_attr}_hash' not in anon_doc
                or anon_doc[f'{auth_attr}_hash'] != Config.anon_token
            ):
                logger.debug(f'Detected change in \'anon_token\' Config Attr.')
                anon_doc_update[attr] = Config.anon_token
            anon_doc_update[f'{auth_attr}_hash'] = Config.anon_token
        if len(anon_doc_update.keys()):
            logger.debug(
                f'Attempting to update ANON user with doc: \'{anon_doc_update}\''
            )
            anon_results = await Config.modules['user'].update(
                skip_events=[Event.PERM, Event.PRE, Event.ON],
                env=Config._sys_env,
                query=[{'_id': ObjectId('f00000000000000000000011')}],
                doc=anon_doc_update,
            )
            logger.debug(f'ANON user update results: {anon_results}')
            if anon_results['status'] != 200:
                logger.error('Config step failed. Exiting.')
                exit(1)
        else:
            logger.debug('ANON user is up-to-date.')

    Config._sys_docs[ObjectId('f00000000000000000000011')] = SYS_DOC(module='user')

    logger.debug('Testing sessions collection.')
    # [DOC] Test if ANON session exists
    session_results = await Config.modules['session'].read(
        skip_events=[Event.PERM, Event.ON],
        env=Config._sys_env,
        query=[{'_id': 'f00000000000000000000012'}],
    )
    if not session_results['args']['count']:
        logger.debug('ANON session not found, creating it.')
        anon_results = await Config.modules['session'].create(
            skip_events=[Event.PERM, Event.PRE, Event.ON],
            env=Config._sys_env,
            doc=_compile_anon_session(),
        )
        logger.debug(f'ANON session creation results: {anon_results}')
        if anon_results['status'] != 200:
            logger.error('Config step failed. Exiting.')
            exit(1)
    Config._sys_docs[ObjectId('f00000000000000000000012')] = SYS_DOC(module='session')

    logger.debug('Testing groups collection.')
    # [DOC] Test if DEFAULT group exists
    group_results = await Config.modules['group'].read(
        skip_events=[Event.PERM, Event.ON],
        env=Config._sys_env,
        query=[{'_id': 'f00000000000000000000013'}],
    )
    if not group_results['args']['count']:
        logger.debug('DEFAULT group not found, creating it.')
        group_create_doc = {
            '_id': ObjectId('f00000000000000000000013'),
            'user': ObjectId('f00000000000000000000010'),
            'name': {locale: '__DEFAULT' for locale in Config.locales},
            'bio': {locale: '__DEFAULT' for locale in Config.locales},
            'privileges': Config.default_privileges,
        }
        group_results = await Config.modules['group'].create(
            skip_events=[Event.PERM, Event.PRE, Event.ON],
            env=Config._sys_env,
            doc=group_create_doc,
        )
        logger.debug(f'DEFAULT group creation results: {group_results}')
        if group_results['status'] != 200:
            logger.error('Config step failed. Exiting.')
            exit(1)
    else:
        logger.debug('DEFAULT group found, checking it.')
        default_doc = group_results['args']['docs'][0]
        default_doc_update: Dict[str, Any] = {}
        for module in Config.default_privileges.keys():
            if module not in default_doc['privileges'].keys() or set(
                default_doc['privileges'][module]
            ) != set(Config.default_privileges[module]):
                logger.debug(f'Detected change in \'default_privileges\' Config Attr.')
                default_doc_update[f'privileges.{module}'] = Config.default_privileges[
                    module
                ]
        if len(default_doc_update.keys()):
            logger.debug(
                f'Attempting to update DEFAULT group with doc: \'{default_doc_update}\''
            )
            default_results = await Config.modules['group'].update(
                skip_events=[Event.PERM, Event.PRE, Event.ON],
                env=Config._sys_env,
                query=[{'_id': ObjectId('f00000000000000000000013')}],
                doc=default_doc_update,
            )
            logger.debug(f'DEFAULT group update results: {default_results}')
            if anon_results['status'] != 200:
                logger.error('Config step failed. Exiting.')
                exit(1)
        else:
            logger.debug('DEFAULT group is up-to-date.')

    Config._sys_docs[ObjectId('f00000000000000000000013')] = SYS_DOC(module='group')

    # [DOC] Test app-specific groups
    logger.debug('Testing app-specific groups collection.')
    for group in Config.groups:
        group_results = await Config.modules['group'].read(
            skip_events=[Event.PERM, Event.ON],
            env=Config._sys_env,
            query=[{'_id': group['_id']}],
        )
        if not group_results['args']['count']:
            logger.debug(
                f'App-specific group with name \'{group["name"]}\' not found, creating it.'
            )
            group_results = await Config.modules['group'].create(
                skip_events=[Event.PERM, Event.PRE, Event.ON],
                env=Config._sys_env,
                doc=group,
            )
            logger.debug(
                f'App-specific group with name {group["name"]} creation results: {group_results}'
            )
            if group_results['status'] != 200:
                logger.error('Config step failed. Exiting.')
                exit(1)
        else:
            logger.debug(
                f'App-specific group with name \'{group["name"]}\' found, checking it.'
            )
            group_doc = group_results['args']['docs'][0]
            group_doc_update = {}
            if 'privileges' in group.keys():
                for module in group['privileges'].keys():
                    if module not in group_doc['privileges'].keys() or set(
                        group_doc['privileges'][module]
                    ) != set(group['privileges'][module]):
                        logger.debug(
                            f'Detected change in \'privileges\' Doc Arg for group with name \'{group["name"]}\'.'
                        )
                        group_doc_update[f'privileges.{module}'] = group['privileges'][
                            module
                        ]
            if len(group_doc_update.keys()):
                logger.debug(
                    f'Attempting to update group with name \'{group["name"]}\' with doc: \'{group_doc_update}\''
                )
                group_results = await Config.modules['group'].update(
                    skip_events=[Event.PERM, Event.PRE, Event.ON],
                    env=Config._sys_env,
                    query=[{'_id': group['_id']}],
                    doc=group_doc_update,
                )
                logger.debug(
                    f'Group with name \'{group["name"]}\' update results: {group_results}'
                )
                if group_results['status'] != 200:
                    logger.error('Config step failed. Exiting.')
                    exit(1)
            else:
                logger.debug(f'Group with name \'{group["name"]}\' is up-to-date.')

        Config._sys_docs[ObjectId(group['_id'])] = SYS_DOC(module='group')

    # [DOC] Test app-specific data indexes
    logger.debug('Testing data indexes')
    for index in Config.data_indexes:
        logger.debug(f'Attempting to create data index: {index}')
        try:
            Config._sys_conn[Config.data_name][index['collection']].create_index(
                index['index']
            )
        except Exception as e:
            logger.error(f'Failed to create data index: {index}, with error: {e}')
            logger.error('Evaluate error and take action manually.')

    logger.debug(
        'Creating \'var\', \'type\', \'user\' data indexes for settings collections.'
    )
    Config._sys_conn[Config.data_name]['settings'].create_index([('var', 1)])
    Config._sys_conn[Config.data_name]['settings'].create_index([('type', 1)])
    Config._sys_conn[Config.data_name]['settings'].create_index([('user', 1)])
    logger.debug(
        'Creating \'user\', \'event\', \'subevent\' data indexes for analytics collections.'
    )
    Config._sys_conn[Config.data_name]['analytics'].create_index([('user', 1)])
    Config._sys_conn[Config.data_name]['analytics'].create_index([('event', 1)])
    Config._sys_conn[Config.data_name]['analytics'].create_index([('subevent', 1)])
    logger.debug('Creating \'__deleted\' data indexes for all collections.')
    for module in Config.modules:
        if Config.modules[module].collection:
            logger.debug(
                f'Attempting to create \'__deleted\' data index for collection: {Config.modules[module].collection}'
            )
            Config._sys_conn[Config.data_name][
                Config.modules[module].collection
            ].create_index([('__deleted', 1)])

    # [DOC] Test app-specific docs
    logger.debug('Testing docs.')
    for doc in Config.docs:
        if type(doc) != SYS_DOC:
            logger.error(f'Invalid Config Attr \'docs\'. Exiting.')
            exit(1)

        doc_results = await Config.modules[doc.module].read(
            skip_events=[Event.PERM, Event.PRE, Event.ON, Event.ARGS],
            env=Config._sys_env,
            query=[{doc.key: doc.key_value}],  # type: ignore
        )
        if not doc_results['args']['count']:
            skip_events = [Event.PERM]
            if doc.skip_args == True:
                skip_events.append(Event.ARGS)
            doc.doc = cast(NAWAH_DOC, doc.doc)
            doc_results = await Config.modules[doc.module].create(
                skip_events=skip_events, env=Config._sys_env, doc=doc.doc
            )
            logger.debug(
                'App-specific doc with %s \'%s\' of module \'%s\' creation results: %s',
                doc.key,
                doc.key_value,
                doc.module,
                doc_results,
            )
            if doc_results['status'] != 200:
                logger.error('Config step failed. Exiting.')
                exit(1)
        Config._sys_docs[ObjectId(doc_results['args']['docs'][0]['_id'])] = SYS_DOC(
            module=doc.module
        )

    # [DOC] Check for emulate_test mode
    if Config.emulate_test:
        Config.test = True


def _compile_anon_user():
    from nawah.utils import generate_attr

    anon_doc = {
        '_id': ObjectId('f00000000000000000000011'),
        'name': {Config.locale: '__ANON'},
        'groups': [],
        'privileges': Config.anon_privileges,
        'locale': Config.locale,
    }
    for attr in Config.user_attrs.keys():
        anon_doc[attr] = generate_attr(attr_type=Config.user_attrs[attr])
    for auth_attr in Config.user_attrs.keys():
        anon_doc[f'{auth_attr}_hash'] = Config.anon_token
    return anon_doc


def _compile_anon_session():
    session_doc = {
        '_id': ObjectId('f00000000000000000000012'),
        'user': ObjectId('f00000000000000000000011'),
        'host_add': '127.0.0.1',
        'user_agent': Config.anon_token,
        'timestamp': '1970-01-01T00:00:00',
        'expiry': '1970-01-01T00:00:00',
        'token': Config.anon_token,
        'token_hash': Config.anon_token,
    }
    return session_doc



import logging
from typing import (TYPE_CHECKING, Any, Dict, List, Literal, Optional,
                    Protocol, Tuple, TypedDict, Union, cast)

from nawah.config import Config
from nawah.enums import NAWAH_VALUES, Event

if TYPE_CHECKING:
    from nawah.base_method import BaseMethod
    from nawah.base_module import BaseModule

    from ._attr import ATTR
    from ._query import Query
    from ._types import NAWAH_DOC, NAWAH_ENV, NAWAH_EVENTS, NAWAH_QUERY


logger = logging.getLogger('nawah')

PRE_HANDLER_RETURN = Tuple[
    'NAWAH_EVENTS', 'NAWAH_ENV', 'Query', 'NAWAH_DOC', Dict[str, Any]
]
ON_HANDLER_RETURN = Tuple[
    Dict[str, Any], 'NAWAH_EVENTS', 'NAWAH_ENV', 'Query', 'NAWAH_DOC', Dict[str, Any]
]

PERM_QUERY_MOD_UNIT = Dict[str, Union['ATTR', Literal['$__date', '$__user'], Any]]
PERM_QUERY_MOD = Union[PERM_QUERY_MOD_UNIT, 'NAWAH_QUERY', List['PERM_QUERY_MOD']]  # type: ignore

PERM_DOC_MOD_UNIT = Dict[str, Union['ATTR', Literal['$__date', '$__user'], Any]]
PERM_DOC_MOD = Union[PERM_DOC_MOD_UNIT, 'NAWAH_DOC']

RESULTS = TypedDict(
    'RESULTS',
    {
        'status': int,
        'msg': str,
        'args': 'RESULTS_ARGS',
    },
)
RESULTS_ARGS = TypedDict(
    'RESULTS_ARGS',
    {
        'code': str,
        'count': int,
        'total': int,
        'docs': List[Dict],
        'session': Dict,
        'cache_key': str,
        'cache_time': str,
    },
)


class PERM:
    privilege: str
    query_mod: PERM_QUERY_MOD
    doc_mod: PERM_DOC_MOD

    _method: 'METHOD'
    _set_index: int

    def __repr__(self):
        return f'<PERM:{self.privilege},{self.query_mod},{self.doc_mod}>'

    def __init__(
        self,
        *,
        privilege: str,
        query_mod: Optional[PERM_QUERY_MOD] = None,
        doc_mod: Optional[PERM_DOC_MOD] = None,
    ):
        if not query_mod:
            query_mod = {}
        if not doc_mod:
            doc_mod = {}
        self.privilege = privilege
        self.query_mod = query_mod
        self.doc_mod = doc_mod

    def _validate(self):
        # [DOC] Validate privilege is listed in BaseModule.privileges to detect typos, missing privileges, but skip *, __sys
        if self.privilege not in ['*', '__sys']:
            # [DOC] Check for cross-module privilege
            if '.' in self.privilege:
                # [DOC] Validate correct format of cross-platform privilege: module_name.privilege
                try:
                    module_name, privilege = self.privilege.split('.')
                except ValueError:
                    raise Exception(
                        f'Permission Set \'{self._set_index}\' of method \'{self._method._method_name}\' of module \'{self._method._module.module_name}\' requires invalid privilege \'{self.privilege}\'.'
                    )

                module = Config.modules[module_name]
            else:
                module_name = self._method._module.module_name
                module = self._method._module
                privilege = self.privilege

            if privilege not in module.privileges:
                raise Exception(
                    f'Permission Set \'{self._set_index}\' of method \'{self._method._method_name}\' of module \'{self._method._module.module_name}\' requires unknown privilege \'{module_name}.{privilege}\'.'
                )

        # [DOC] Add default Doc Modifiers to prevent sys attrs from being modified
        if self._method._method_name == 'update':
            for attr in ['user', 'create_time']:
                if attr not in self.doc_mod.keys():
                    self.doc_mod[attr] = None
                # [TODO] Assert this is behaving correctly
                elif self.doc_mod[attr] == NAWAH_VALUES.ALLOW_MOD:
                    del self.doc_mod[attr]

        # [DOC] Validate query_mod, doc_mod
        self._validate_query_mod()
        self._validate_doc_mod()

    def _validate_query_mod(self):
        if self.query_mod:
            logger.debug(
                f'Attemting to validate \'{self._method._module.module_name}.{self._method._method_name}[{self._set_index}].query_mod\'.'
            )
            self.__validate_obj(self.query_mod)

    def _validate_doc_mod(self):
        if self.doc_mod:
            logger.debug(
                f'Attemting to validate \'{self._method._module.module_name}.{self._method._method_name}[{self._set_index}].doc_mod\'.'
            )
            self.__validate_obj(self.doc_mod)

    def __validate_obj(self, obj):
        from ._attr import ATTR

        obj_iter: Iterable
        if type(obj) == list:
            obj_iter = range(len(obj))
        elif type(obj) == dict:
            obj_iter = list(obj.keys())
        else:
            raise Exception(
                f'Can\'t validate permission obj {obj} of type {type(obj)}.'
            )

        logger.debug(f'Found \'{obj}\' to validate. Iterating.')

        for j in obj_iter:
            logger.debug(f'Attempting to validate \'{j}\'')
            if type(obj[j]) in [list, dict]:
                logger.debug(f'Item is an iterable object. Iterating.')
                self.__validate_obj(obj[j])
            elif type(obj[j]) == ATTR:
                if obj[j]._type != 'TYPE':
                    raise Exception(
                        f'Attr Type \'{j}\' of TYPE \'{obj[j]._type}\' is not allowed.'
                    )

                ATTR.validate_type(attr_type=obj[j])


class EXTN:
    module: str
    skip_events: Optional[List[Event]]
    query: Optional['NAWAH_QUERY']
    attrs: Union[List[str], str]
    force: Union[bool, str] = False

    def __repr__(self):
        return f'<EXTN:{self.module},{self.attrs},{self.force}>'

    def __init__(
        self,
        *,
        module: str,
        skip_events: List[Event] = None,
        query: 'NAWAH_QUERY' = None,
        attrs: Union[List[str], str],
        force: Union[bool, str] = False,
    ):
        self.module = module
        self.skip_events = skip_events
        self.query = query
        self.attrs = attrs
        self.force = force

        # [DOC] Wrap query in list if it is a dict
        if type(query) == dict:
            self.query = [self.query]


class METHOD:
    permissions: List[PERM]
    query_args: Union[
        None,
        List[Dict[str, 'ATTR']],
        Dict[str, 'ATTR'],
    ]
    doc_args: Union[
        None,
        List[Dict[str, 'ATTR']],
        Dict[str, 'ATTR'],
    ] = None
    get_method: bool
    post_method: bool
    watch_method: bool
    cache: Union[None, 'CACHE'] = None
    reset_cache_channels: Union[None, List[str]] = None

    _callable: 'BaseMethod'
    _module: 'BaseModule'
    _method_name: str

    def __init__(
        self,
        *,
        permissions: List[PERM],
        query_args: Union[
            None,
            List[Dict[str, 'ATTR']],
            Dict[str, 'ATTR'],
        ] = None,
        doc_args: Union[
            None,
            List[Dict[str, 'ATTR']],
            Dict[str, 'ATTR'],
        ] = None,
        get_method: bool = False,
        post_method: bool = False,
        watch_method: bool = False,
        cache: Union[None, 'CACHE'] = None,
        reset_cache_channels: Union[None, List[str]] = None,
    ):
        self.permissions = permissions
        self.query_args = query_args
        self.doc_args = doc_args
        self.get_method = get_method
        self.post_method = post_method
        self.watch_method = watch_method
        self.cache = cache
        self.reset_cache_channels = reset_cache_channels

    def __call__(self, **kwargs):
        return self._callable(**kwargs)

    def _validate(self):
        from nawah.base_method import BaseMethod

        from ._attr import ATTR

        # [DOC] Check for existence of at least single permissions set per method
        if not len(self.permissions):
            raise Exception(
                f'No permissions sets for method \'{self._method_name}\' of module \'{self._module.module_name}\'.'
            )
        # [DOC] Check method query_args attr, set it or update it if required.
        if not self.query_args:
            if self._method_name == 'create_file':
                self.query_args = [{'_id': ATTR.ID(), 'attr': ATTR.STR()}]
            elif self._method_name == 'delete_file':
                self.query_args = [
                    {
                        '_id': ATTR.ID(),
                        'attr': ATTR.STR(),
                        'index': ATTR.INT(),
                        'name': ATTR.STR(),
                    }
                ]
        elif type(self.query_args) == dict:
            method_query_args = self.query_args
            method_query_args = cast(Dict[str, ATTR], method_query_args)
            self.query_args = [method_query_args]
        # [DOC] Check method doc_args attr, set it or update it if required.
        if not self.doc_args:
            if self._method_name == 'create_file':
                self.doc_args = [{'file': ATTR.FILE()}]
        elif type(self.doc_args) == dict:
            method_doc_args = self.doc_args
            method_doc_args = cast(Dict[str, ATTR], method_doc_args)
            self.doc_args = [method_doc_args]
        # [DOC] Check method get_method attr, update it if required.
        if self.get_method == True:
            if not self.query_args:
                if self._method_name == 'retrieve_file':
                    self.query_args = [
                        {
                            '_id': ATTR.ID(),
                            'attr': ATTR.STR(),
                            'filename': ATTR.STR(),
                        },
                        {
                            '_id': ATTR.ID(),
                            'attr': ATTR.STR(),
                            'thumb': ATTR.STR(pattern=r'[0-9]+x[0-9]+'),
                            'filename': ATTR.STR(),
                        },
                    ]
                else:
                    self.query_args = [{}]
        # [DOC] Check method post_method attr, update it if required.
        if self.post_method == True:
            if not self.query_args:
                self.query_args = [{}]
        # [DOC] Check permissions sets for any invalid set
        for i in range(len(self.permissions)):
            permissions_set = self.permissions[i]
            if type(permissions_set) != PERM:
                raise Exception(
                    f'Invalid permissions set \'{permissions_set}\' of method \'{self._method_name}\' of module \'{self._module.module_name}\'.'
                )
            # [DOC] Set PERM._method value to create back-link from child to parent
            permissions_set._method = self
            permissions_set._set_index = i
            permissions_set = cast(PERM, permissions_set)
            # [DOC] Check valida Permission Set
            permissions_set._validate()

        # [DOC] Check invalid query_args, doc_args types
        for arg_set in ['query_args', 'doc_args']:
            arg_set = cast(Literal['query_args', 'doc_args'], arg_set)
            if getattr(self, arg_set):
                method_arg_set: List[Dict[str, ATTR]] = getattr(self, arg_set)
                for args_set in method_arg_set:
                    for attr in args_set.keys():
                        try:
                            ATTR.validate_type(attr_type=args_set[attr])
                        except:
                            raise Exception(
                                f'Invalid \'{arg_set}\' attr type for \'{attr}\' of set \'{args_set}\' of method \'{self._method_name}\' of module \'{self._module.module_name}\'.'
                            )
        # [DOC] Initialise method as BaseMethod
        method_query_args = self.query_args  # type: ignore
        method_query_args = cast(List[Dict[str, ATTR]], method_query_args)
        method_doc_args = self.doc_args  # type: ignore
        method_doc_args = cast(List[Dict[str, ATTR]], method_doc_args)

        # [DOC] Validate method implementation exist in the class
        try:
            getattr(self._module, f'_method_{self._method_name}')
        except AttributeError:
            raise Exception(
                f'Method \'{self._method_name}\' of module \'{self._module.module_name}\' is defined but not implemented.'
            )

        self._callable = BaseMethod(
            module=self._module,
            method=self._method_name,
            permissions=self.permissions,
            query_args=method_query_args,
            doc_args=method_doc_args,
            watch_method=self.watch_method,
            get_method=self.get_method,
            post_method=self.post_method,
            cache=self.cache,
            reset_cache_channels=self.reset_cache_channels,
        )


class CACHE_CONDITION(Protocol):
    def __call__(
        self,
        skip_events: 'NAWAH_EVENTS',
        env: 'NAWAH_ENV',
        query: Union['Query', 'NAWAH_QUERY'],
    ) -> bool:
        ...


class CACHE:
    channel: str
    condition: CACHE_CONDITION
    user_scoped: bool
    period: Optional[int]

    def __repr__(self):
        return f'<CACHE:{self.condition},{self.user_scoped},{self.period}>'

    def __init__(
        self,
        *,
        channel: str,
        condition: CACHE_CONDITION,
        user_scoped: bool,
        period: int = None,
    ):
        self.channel = channel
        self.condition = condition
        self.user_scoped = user_scoped
        self.period = period


class ANALYTIC_CONDITION(Protocol):
    def __call__(
        self,
        skip_events: 'NAWAH_EVENTS',
        env: 'NAWAH_ENV',
        query: Union['Query', 'NAWAH_QUERY'],
        doc: 'NAWAH_DOC',
        method: str,
    ) -> bool:
        ...


class ANALYTIC_DOC(Protocol):
    def __call__(
        self,
        skip_events: 'NAWAH_EVENTS',
        env: 'NAWAH_ENV',
        query: Union['Query', 'NAWAH_QUERY'],
        doc: 'NAWAH_DOC',
        method: str,
    ) -> 'NAWAH_DOC':
        ...


class ANALYTIC:
    condition: ANALYTIC_CONDITION
    doc: ANALYTIC_DOC

    def __init__(
        self,
        *,
        condition: ANALYTIC_CONDITION,
        doc: ANALYTIC_DOC,
    ):
        setattr(self, 'condition', condition)
        setattr(self, 'doc', doc)



import datetime

from nawah.base_module import BaseModule
from nawah.classes import ATTR, METHOD, PERM
from nawah.enums import Event


class Analytic(BaseModule):
    '''`Analytic` module provides data type and controller from `Analytics Workflow` and accompanying analytics docs. It uses `pre_create` handler to assure no events duplications occur and all occurrences of the same event are recorded in one doc.'''

    collection = 'analytics'
    attrs = {
        'user': ATTR.ID(desc='`_id` of `User` doc the doc belongs to.'),
        'event': ATTR.STR(desc='Analytics event name.'),
        'subevent': ATTR.ANY(
            desc='Analytics subevent distinguishing attribute. This is usually `STR`, or `ID` but it is introduced in the module as `ANY` to allow wider use-cases by developers.'
        ),
        'date': ATTR.DATE(
            desc='Analytics event date. This allows clustering of events occupancies to limit doc size.'
        ),
        'occurrences': ATTR.LIST(
            desc='All occurrences of the event as list.',
            list=[
                ATTR.TYPED_DICT(
                    desc='Single occurrence of the event details.',
                    dict={
                        'args': ATTR.KV_DICT(
                            desc='Key-value `dict` containing event args, if any.',
                            key=ATTR.STR(),
                            val=ATTR.ANY(),
                        ),
                        'score': ATTR.INT(
                            desc='Numerical score for occurrence of the event.'
                        ),
                        'create_time': ATTR.DATETIME(
                            desc='Python `datetime` ISO format of the occurrence of the event.'
                        ),
                    },
                )
            ],
        ),
        'score': ATTR.INT(
            desc='Total score of all scores of all occurrences of the event. This can be used for data analysis.'
        ),
    }
    unique_attrs = [('user', 'event', 'subevent', 'date')]
    methods = {
        'read': METHOD(permissions=[PERM(privilege='read')]),
        'create': METHOD(
            permissions=[PERM(privilege='__sys')],
            doc_args={
                'event': ATTR.STR(),
                'subevent': ATTR.ANY(),
                'args': ATTR.KV_DICT(key=ATTR.STR(), val=ATTR.ANY()),
            },
        ),
        'update': METHOD(permissions=[PERM(privilege='__sys')]),
        'delete': METHOD(permissions=[PERM(privilege='delete')]),
    }

    async def pre_create(self, skip_events, env, query, doc, payload):
        if 'session' not in env:
            results = self.status(
                status=200,
                msg='Skipped Analytics Workflow due to None-value `env[session]`.',
            )
            return (skip_events, env, query, doc, {'__results': results})

        analytic_results = await self.read(
            skip_events=[Event.PERM],
            env=env,
            query=[
                {
                    'user': env['session']['user']['_id'],
                    'event': doc['event'],
                    'subevent': doc['subevent'],
                    'date': datetime.date.today().isoformat(),
                },
                {'$limit': 1},
            ],
        )
        if analytic_results['args']['count']:
            analytic_results = await self.update(
                skip_events=[Event.PERM],
                env=env,
                query=[{'_id': analytic_results['args']['docs'][0]['_id']}],
                doc={
                    'occurrences': {
                        '$append': {
                            'args': doc['args'],
                            'score': doc['score'] if 'score' in doc.keys() else 0,
                            'create_time': datetime.datetime.utcnow().isoformat(),
                        }
                    },
                    'score': {'$add': doc['score'] if 'score' in doc.keys() else 0},
                },
            )
            if analytic_results['status'] == 200:
                return (skip_events, env, query, doc, {'__results': analytic_results})

            raise self.exception(
                status=analytic_results['status'],
                msg=analytic_results['msg'],
                args=analytic_results['args'],
            )
        else:
            doc = {
                'event': doc['event'],
                'subevent': doc['subevent'],
                'date': datetime.date.today().isoformat(),
                'occurrences': [
                    {
                        'args': doc['args'],
                        'score': doc['score'] if 'score' in doc.keys() else 0,
                        'create_time': datetime.datetime.utcnow().isoformat(),
                    }
                ],
                'score': doc['score'] if 'score' in doc.keys() else 0,
            }
            return (skip_events, env, query, doc, payload)



from nawah.config import Config
from motor.motor_asyncio import AsyncIOMotorClient
from typing import Dict, Any

import logging, os.path

logger = logging.getLogger('nawah')


def create_conn() -> AsyncIOMotorClient:
    connection_config: Dict[str, Any] = {'ssl': Config.data_ssl}
    if Config.data_ca and Config.data_ca_name:
        __location__ = os.path.realpath(
            os.path.join(os.getcwd(), os.path.dirname(__file__))
        )
        connection_config['ssl_ca_certs'] = os.path.join(
            __location__, '..', 'certs', Config.data_ca_name
        )
    # [DOC] Check for multiple servers
    if type(Config.data_server) == list:
        for data_server in Config.data_server:
            conn = AsyncIOMotorClient(data_server, **connection_config, connect=True)
            try:
                logger.debug(f'Check if data_server: {data_server} isMaster.')
                results = conn.admin.command('ismaster')
                logger.debug(f'-Check results: {results}')
                if results['ismaster']:
                    break
            except Exception as err:
                logger.debug(f'Not master. Error: {err}')
                pass
    elif type(Config.data_server) == str:
        # [DOC] If it's single server just connect directly
        conn = AsyncIOMotorClient(Config.data_server, **connection_config, connect=True)
    return conn



from nawah.enums import NAWAH_VALUES
from nawah.config import Config

from bson import ObjectId
from typing import (
    Dict,
    Union,
    Type,
    List,
    Sequence,
    Any,
    Literal,
    Callable,
    Optional,
    ForwardRef,
    TYPE_CHECKING,
    Protocol,
)

import datetime, logging, re, inspect

from ._package import SYS_DOC
from ._exceptions import (
    InvalidAttrTypeException,
    InvalidAttrTypeArgException,
    InvalidAttrTypeArgsException,
)

if TYPE_CHECKING:
    from ._module import EXTN
    from ._types import NAWAH_EVENTS, NAWAH_ENV, NAWAH_QUERY, NAWAH_DOC, NAWAH_DOC
    from ._query import Query

logger = logging.getLogger('nawah')

ATTRS_TYPES_ARGS: Dict[str, Dict[str, Union[Type, str]]] = {
    'ANY': {},
    'ACCESS': {},
    'COUNTER': {'pattern': str},
    'ID': {},
    'STR': {'pattern': str},
    'INT': {'ranges': List[List[int]]},
    'FLOAT': {'ranges': List[List[float]]},
    'BOOL': {},
    'LOCALE': {},
    'LOCALES': {},
    'EMAIL': {
        'allowed_domains': List[str],
        'disallowed_domains': List[str],
        'strict_matching': bool,
    },
    'PHONE': {'codes': List[str]},
    'IP': {},
    'URI_WEB': {
        'allowed_domains': List[str],
        'disallowed_domains': List[str],
        'strict_matching': bool,
    },
    'DATETIME': {'ranges': List[List[datetime.datetime]]},
    'DATE': {'ranges': List[List[datetime.date]]},
    'DYNAMIC_ATTR': {'types': List[str]},
    'DYNAMIC_VAL': {'dynamic_attr': str},
    'TIME': {'ranges': List[List[datetime.time]]},
    'FILE': {
        'types': List[str],
        'max_ratio': List[int],
        'min_ratio': List[int],
        'max_dims': List[int],
        'min_dims': List[int],
        'max_size': int,
    },
    'GEO': {},
    'LIST': {'list': List['ATTR'], 'min': int, 'max': int},
    'KV_DICT': {
        'key': 'ATTR',
        'val': 'ATTR',
        'min': int,
        'max': int,
        'req': List[str],
    },
    'TYPED_DICT': {'dict': Dict[str, 'ATTR']},
    'LITERAL': {'literal': List[Union[str, int, float, bool]]},
    'UNION': {'union': List['ATTR']},
    'TYPE': {'type': str},
}

SPECIAL_ATTRS = Literal[
    '$search',
    '$sort',
    '$skip',
    '$limit',
    '$extn',
    '$soft',
    '$attrs',
    '$group',
    '$geo_near',
]


class ATTR_TYPE_CALLABLE_TYPE(Protocol):
    def __call__(
        *,
        self,
        mode: Literal['create', 'create_draft', 'update'],
        attr_name: str,
        attr_type: 'ATTR',
        attr_val: Any,
        skip_events: 'NAWAH_EVENTS',
        env: 'NAWAH_ENV',
        query: Union['NAWAH_QUERY', 'Query'],
        doc: 'NAWAH_DOC',
        scope: Optional['NAWAH_DOC'],
    ) -> Any:
        ...


ATTRS_TYPES_TYPE = Literal[
    'ANY',
    'ACCESS',
    'COUNTER',
    'ID',
    'STR',
    'INT',
    'FLOAT',
    'BOOL',
    'LOCALE',
    'LOCALES',
    'EMAIL',
    'PHONE',
    'IP',
    'URI_WEB',
    'DATETIME',
    'DATE',
    'DYNAMIC_ATTR',
    'DYNAMIC_VAL',
    'TIME',
    'FILE',
    'GEO',
    'LIST',
    'KV_DICT',
    'TYPED_DICT',
    'LITERAL',
    'UNION',
    'TYPE',
]


class ATTR:
    _nawah_attr: bool = True
    _type: ATTRS_TYPES_TYPE
    _desc: Optional[str]
    _args: Dict[str, Any]
    _valid: bool = False
    _extn: Optional[Union['EXTN', 'ATTR']] = None

    __default = NAWAH_VALUES.NONE_VALUE

    @property
    def _default(self):
        if self.__default == '$__datetime':
            return datetime.datetime.utcnow().isoformat()
        elif self.__default == '$__date':
            return datetime.date.today().isoformat()
        elif self.__default == '$__time':
            return datetime.datetime.now().time().isoformat()
        else:
            return self.__default

    @_default.setter
    def _default(self, value):
        self.__default = value

    def __repr__(self):
        return f'<ATTR:{self._type},{self._args}>'

    def __init__(self, *, attr_type: ATTRS_TYPES_TYPE, desc: str = None, **kwargs: Any):
        self._type = attr_type
        self._desc = desc
        self._args = kwargs
        ATTR.validate_type(attr_type=self, skip_type=True)

    @classmethod
    def ANY(cls, *, desc: str = None):
        return ATTR(attr_type='ANY', desc=desc)

    @classmethod
    def ACCESS(cls, *, desc: str = None):
        return ATTR(attr_type='ACCESS', desc=desc)

    @classmethod
    def COUNTER(cls, *, desc: str = None, pattern: str, values: List[Callable] = None):
        return ATTR(attr_type='COUNTER', desc=desc, pattern=pattern, values=values)

    @classmethod
    def ID(cls, *, desc: str = None):
        return ATTR(attr_type='ID', desc=desc)

    @classmethod
    def STR(cls, *, desc: str = None, pattern: str = None):
        return ATTR(attr_type='STR', desc=desc, pattern=pattern)

    @classmethod
    def INT(cls, *, desc: str = None, ranges: List[List[int]] = None):
        return ATTR(attr_type='INT', desc=desc, ranges=ranges)

    @classmethod
    def FLOAT(cls, *, desc: str = None, ranges: List[List[int]] = None):
        return ATTR(attr_type='FLOAT', desc=desc, ranges=ranges)

    @classmethod
    def BOOL(cls, *, desc: str = None):
        return ATTR(attr_type='BOOL', desc=desc)

    @classmethod
    def LOCALE(cls, *, desc: str = None):
        return ATTR(attr_type='LOCALE', desc=desc)

    @classmethod
    def LOCALES(cls, *, desc: str = None):
        return ATTR(attr_type='LOCALES', desc=desc)

    @classmethod
    def EMAIL(
        cls,
        *,
        desc: str = None,
        allowed_domains: List[str] = None,
        disallowed_domains: List[str] = None,
        strict: bool = False,
    ):
        return ATTR(
            attr_type='EMAIL',
            desc=desc,
            allowed_domains=allowed_domains,
            disallowed_domains=disallowed_domains,
            strict=strict,
        )

    @classmethod
    def PHONE(cls, *, desc: str = None, codes: List[str] = None):
        return ATTR(attr_type='PHONE', desc=desc, codes=codes)

    @classmethod
    def IP(cls, *, desc: str = None):
        return ATTR(attr_type='IP', desc=desc)

    @classmethod
    def URI_WEB(
        cls,
        *,
        desc: str = None,
        allowed_domains: List[str] = None,
        disallowed_domains: List[str] = None,
        strict: bool = False,
    ):
        return ATTR(
            attr_type='URI_WEB',
            desc=desc,
            allowed_domains=allowed_domains,
            disallowed_domains=disallowed_domains,
            strict=strict,
        )

    @classmethod
    def DATETIME(cls, *, desc: str = None, ranges: List[List[str]] = None):
        return ATTR(attr_type='DATETIME', desc=desc, ranges=ranges)

    @classmethod
    def DATE(cls, *, desc: str = None, ranges: List[List[str]] = None):
        return ATTR(attr_type='DATE', desc=desc, ranges=ranges)

    @classmethod
    def DYNAMIC_ATTR(cls, *, desc: str = None, types: List[str] = None):
        return ATTR(attr_type='DYNAMIC_ATTR', desc=desc, types=types)

    @classmethod
    def DYNAMIC_VAL(cls, *, desc: str = None, dynamic_attr: str):
        return ATTR(attr_type='DYNAMIC_VAL', desc=desc, dynamic_attr=dynamic_attr)

    @classmethod
    def TIME(cls, *, desc: str = None, ranges: List[List[str]] = None):
        return ATTR(attr_type='TIME', desc=desc, ranges=ranges)

    @classmethod
    def FILE(
        cls,
        *,
        desc: str = None,
        types: List[str] = None,
        max_ratio: List[int] = None,
        min_ratio: List[int] = None,
        max_dims: List[int] = None,
        min_dims: List[int] = None,
        max_size: int = None,
    ):
        return ATTR(
            attr_type='FILE',
            desc=desc,
            types=types,
            max_ratio=max_ratio,
            min_ratio=min_ratio,
            max_dims=max_dims,
            min_dims=min_dims,
            max_size=max_size,
        )

    @classmethod
    def GEO(cls, *, desc: str = None):
        return ATTR(attr_type='GEO', desc=desc)

    @classmethod
    def LIST(
        cls,
        *,
        desc: str = None,
        list: List['ATTR'],
        min: int = None,
        max: int = None,
    ):
        return ATTR(attr_type='LIST', desc=desc, list=list, min=min, max=max)

    @classmethod
    def KV_DICT(
        cls,
        *,
        desc: str = None,
        key: 'ATTR',
        val: 'ATTR',
        min: int = None,
        max: int = None,
        req: List[str] = None,
    ):
        return ATTR(
            attr_type='KV_DICT', desc=desc, key=key, val=val, min=min, max=max, req=req
        )

    @classmethod
    def TYPED_DICT(cls, *, desc: str = None, dict: Dict[str, 'ATTR']):
        return ATTR(attr_type='TYPED_DICT', desc=desc, dict=dict)

    @classmethod
    def LITERAL(
        cls, *, desc: str = None, literal: Sequence[Union[str, int, float, bool]]
    ):
        return ATTR(attr_type='LITERAL', desc=desc, literal=literal)

    @classmethod
    def UNION(cls, *, desc: str = None, union: List['ATTR']):
        return ATTR(attr_type='UNION', desc=desc, union=union)

    @classmethod
    def TYPE(cls, *, desc: str = None, type: Union[str, ATTR_TYPE_CALLABLE_TYPE]):
        return ATTR(attr_type='TYPE', desc=desc, type=type)

    @classmethod
    def validate_type(cls, *, attr_type: 'ATTR', skip_type: bool = False):
        # [DOC] Skip validating Attr Type if it is already validated, unless we want to validate TYPE, which could require deep (nested) validation
        if skip_type and attr_type._valid:
            return

        if attr_type._type not in ATTRS_TYPES_ARGS.keys():
            raise InvalidAttrTypeException(attr_type=attr_type)
        elif not skip_type and attr_type._type == 'TYPE':
            # [DOC] Complex condition is unreadblae in single condition set, break into multiple conditions as try..except block
            try:
                if type(attr_type._args['type']) == str:
                    if attr_type._args['type'] not in Config.types.keys():
                        raise Exception()
                    if not inspect.iscoroutinefunction(
                        Config.types[attr_type._args['type']]
                    ):
                        raise Exception()
                    # [DOC] Assign new Attr Type Arg for shorthand calling the TYPE function
                    attr_type._args['func'] = Config.types[attr_type._args['type']]
                else:
                    if not inspect.iscoroutinefunction(attr_type._args['type']):
                        raise Exception()
                    # [DOC] Assign new Attr Type Arg for shorthand calling the TYPE function
                    attr_type._args['func'] = attr_type._args['type']
            except:
                raise InvalidAttrTypeException(attr_type=attr_type)
        elif attr_type._type != 'TYPE':
            for arg in ATTRS_TYPES_ARGS[attr_type._type].keys():
                if (
                    arg in ['list', 'dict', 'literal', 'union', 'type']
                    and arg not in attr_type._args.keys()
                ):
                    raise InvalidAttrTypeArgException(
                        arg_name=arg,
                        arg_type=ATTRS_TYPES_ARGS[attr_type._type][arg],
                        arg_val=attr_type._args[arg],
                    )
                elif arg in attr_type._args.keys() and attr_type._args[arg] != None:
                    cls.validate_arg(
                        arg_name=arg,
                        arg_type=ATTRS_TYPES_ARGS[attr_type._type][arg],
                        arg_val=attr_type._args[arg],
                        skip_type=skip_type,
                    )
                else:
                    attr_type._args[arg] = None
            if attr_type._type == 'COUNTER':
                if (
                    '$__values.' in attr_type._args['pattern']
                    or '$__counters:' in attr_type._args['pattern']
                ):
                    logger.error(
                        'Attr Type COUNTER is using wrong format for \'$__values\', or \'$__counters\'.'
                    )
                    raise InvalidAttrTypeException(attr_type=attr_type)
                counter_groups = re.findall(
                    r'(\$__(?:values:[0-9]+|counters\.[a-z0-9_]+))',
                    attr_type._args['pattern'],
                )
                if len(counter_groups) == 0:
                    logger.error(
                        'Attr Type COUNTER is not having any \'$__values\', or \'$__counters\'.'
                    )
                    raise InvalidAttrTypeException(attr_type=attr_type)
                if '$__counters.' not in attr_type._args['pattern']:
                    logger.warning(
                        'Attr Type COUNTER is defined with not \'$__counters\'.'
                    )
                for group in counter_groups:
                    if group.startswith('$__counters.'):
                        Config.docs.append(
                            SYS_DOC(
                                module='setting',
                                key='var',
                                doc={
                                    'user': ObjectId('f00000000000000000000010'),
                                    'var': '__counter:'
                                    + group.replace('$__counters.', ''),
                                    'val_type': {
                                        'type': 'INT',
                                        'args': {},
                                        'allow_none': False,
                                        'default': None,
                                    },
                                    'val': 0,
                                    'type': 'global',
                                },
                            )
                        )
            attr_type._valid = True

    @classmethod
    def validate_arg(
        cls, *, arg_name: str, arg_type: Any, arg_val: Any, skip_type: bool
    ):
        if arg_type == str:
            if type(arg_val) != str:
                raise InvalidAttrTypeArgException(
                    arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                )
            return
        elif arg_type == int:
            if type(arg_val) != int:
                raise InvalidAttrTypeArgException(
                    arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                )
            return
        elif arg_type == float:
            if type(arg_val) not in [float, int]:
                raise InvalidAttrTypeArgException(
                    arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                )
            return
        elif arg_type == bool:
            if type(arg_val) != bool:
                raise InvalidAttrTypeArgException(
                    arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                )
            return
        elif type(arg_type) == ForwardRef or (
            type(arg_type) == str and arg_type == 'ATTR'
        ):
            if type(arg_val) != ATTR:
                raise InvalidAttrTypeArgException(
                    arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                )
            if arg_name == 'key':
                if arg_val._type not in ['STR', 'LITERAL']:
                    raise InvalidAttrTypeArgException(
                        arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                    )
            return
        elif arg_name == 'literal':
            if type(arg_val) != list:
                raise InvalidAttrTypeArgException(
                    arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                )
            for arg_val_child in arg_val:
                if type(arg_val_child) not in arg_type.__args__[0].__args__:
                    raise InvalidAttrTypeArgException(
                        arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                    )
            return
        elif arg_name == 'union':
            if type(arg_val) != list:
                raise InvalidAttrTypeArgException(
                    arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                )
            for arg_val_child in arg_val:
                if type(arg_val_child) != ATTR:
                    raise InvalidAttrTypeArgException(
                        arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                    )
                ATTR.validate_type(attr_type=arg_val_child, skip_type=skip_type)
            return
        elif arg_type == datetime.date:
            if not re.match(r'^[0-9]{4}-[0-9]{2}-[0-9]{2}$', arg_val) and not re.match(
                r'^[\-\+][0-9]+[dsmhw]$', arg_val
            ):
                raise InvalidAttrTypeArgException(
                    arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                )
            return
        elif arg_type == datetime.datetime:
            if (
                not re.match(
                    r'^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}(:[0-9]{2}(\.[0-9]{6})?)?$',
                    arg_val,
                )
                and not re.match(r'^[\-\+][0-9]+[dsmhw]$', arg_val)
            ):
                raise InvalidAttrTypeArgException(
                    arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                )
            return
        elif arg_type == datetime.time:
            if not re.match(
                r'^[0-9]{2}:[0-9]{2}(:[0-9]{2}(\.[0-9]{6})?)?$', arg_val
            ) and not re.match(r'^[\-\+][0-9]+[dsmhw]$', arg_val):
                raise InvalidAttrTypeArgException(
                    arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                )
            return
        elif arg_type._name == 'List':
            if type(arg_val) != list:
                raise InvalidAttrTypeArgException(
                    arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                )
            for arg_val_child in arg_val:
                cls.validate_arg(
                    arg_name=arg_name,
                    arg_type=arg_type.__args__[0],
                    arg_val=arg_val_child,
                    skip_type=skip_type,
                )

                # [DOC] In Addition to validating as list arg, check for ATTR and validate
                if type(arg_val_child) == ATTR:
                    ATTR.validate_type(attr_type=arg_val_child, skip_type=skip_type)

            return
        elif arg_type._name == 'Dict':
            if type(arg_val) != dict:
                raise InvalidAttrTypeArgException(
                    arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                )
            for arg_val_child in arg_val.keys():
                if type(arg_val[arg_val_child]) != ATTR:
                    raise InvalidAttrTypeArgException(
                        arg_name=arg_name,
                        arg_type=arg_type,
                        arg_val=arg_val[arg_val_child],
                    )
                ATTR.validate_type(
                    attr_type=arg_val[arg_val_child], skip_type=skip_type
                )
            return

        raise InvalidAttrTypeArgException(
            arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
        )



import copy
import datetime
import logging
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Union

from nawah.classes import (ATTR, NAWAH_DOC, NAWAH_ENV, NAWAH_QUERY, PERM,
                           InvalidAttrException, InvalidPermissionsExcpetion,
                           Query)
from nawah.enums import Event
from nawah.registry import Registry
from nawah.utils import _extract_attr, validate_attr

if TYPE_CHECKING:
    from nawah.base_module import BaseModule

logger = logging.getLogger('nawah')


async def _check_permissions(
    skip_events: List[Event],
    env: NAWAH_ENV,
    query: Union[NAWAH_QUERY, Query],
    doc: NAWAH_DOC,
    module: 'BaseModule',
    permissions: List[PERM],
):
    user = env['session']['user']

    permissions = copy.deepcopy(permissions)

    for permission in permissions:
        logger.debug(f'checking permission: {permission} against: {user["privileges"]}')
        permission_pass = False
        if permission.privilege == '*':
            permission_pass = True

        if not permission_pass:
            if permission.privilege.find('.') == -1:
                permission_module = module.module_name
                permission_attr = permission.privilege
            elif permission.privilege.find('.') != -1:
                permission_module = permission.privilege.split('.')[0]
                permission_attr = permission.privilege.split('.')[1]

            if '*' in user['privileges'].keys():
                user['privileges'][permission_module] = copy.deepcopy(
                    user['privileges']['*']
                )
            if permission_module in user['privileges'].keys():
                if (
                    type(user['privileges'][permission_module]) == list
                    and '*' in user['privileges'][permission_module]
                ):
                    user['privileges'][permission_module] += copy.deepcopy(
                        Registry.module(permission_module).privileges
                    )
            if permission_module not in user['privileges'].keys():
                user['privileges'][permission_module] = []

            if permission_attr in user['privileges'][permission_module]:
                permission_pass = True

        if permission_pass:
            query_mod = await _parse_permission_args(
                skip_events=skip_events,
                env=env,
                query=query,
                doc=doc,
                permission_args=permission.query_mod,
            )
            doc_mod = await _parse_permission_args(
                skip_events=skip_events,
                env=env,
                query=query,
                doc=doc,
                permission_args=permission.doc_mod,
            )

            return {'query_mod': query_mod, 'doc_mod': doc_mod}
    # [DOC] If all permission checks fail
    raise InvalidPermissionsExcpetion()


async def _parse_permission_args(
    skip_events: List[Event],
    env: NAWAH_ENV,
    query: Union[NAWAH_QUERY, Query],
    doc: NAWAH_DOC,
    permission_args: Any,
):
    user = env['session']['user']

    args_iter: Iterable

    if type(permission_args) == list:
        args_iter = range(len(permission_args))
    elif type(permission_args) == dict:
        args_iter = list(permission_args.keys())

    for j in args_iter:
        if type(permission_args[j]) == ATTR:
            try:
                permission_args[j] = await validate_attr(
                    mode='create',
                    attr_name=j,
                    attr_type=permission_args[j],
                    attr_val=doc[j],
                    skip_events=[],
                    env=env,
                    query=query,
                    doc=doc,
                    scope=doc,
                )
            except InvalidAttrException as e:
                raise e
            except:
                # [DOC] There is a chance doc[j] is invalid, so try to get the type in try..except
                try:
                    raise InvalidAttrException(
                        attr_name=j,
                        attr_type=permission_args[j],
                        val_type=type(doc[j]),
                    )
                except:
                    raise InvalidAttrException(
                        attr_name=j,
                        attr_type=permission_args[j],
                        val_type=None,
                    )
        elif type(permission_args[j]) == dict:
            # [DOC] Check opers
            for oper in [
                '$gt',
                '$lt',
                '$gte',
                '$lte',
                '$bet',
                '$ne',
                '$regex',
                '$all',
                '$in',
                '$nin',
            ]:
                if oper in permission_args[j].keys():
                    if oper == '$bet':
                        permission_args[j]['$bet'] = await _parse_permission_args(
                            skip_events=skip_events,
                            env=env,
                            query=query,
                            doc=doc,
                            permission_args=permission_args[j]['$bet'],
                        )
                    else:
                        permission_args[j][oper] = await _parse_permission_args(
                            skip_events=skip_events,
                            env=env,
                            query=query,
                            doc=doc,
                            permission_args={oper: permission_args[j][oper]},
                        )
                        permission_args[j][oper] = permission_args[j][oper][oper]
                    # [DOC] Continue the iteration
                    continue
            # [DOC] Child args, parse
            permission_args[j] = await _parse_permission_args(
                skip_events=skip_events,
                env=env,
                query=query,
                doc=doc,
                permission_args=permission_args[j],
            )
        elif type(permission_args[j]) == list:
            permission_args[j] = await _parse_permission_args(
                skip_events=skip_events,
                env=env,
                query=query,
                doc=doc,
                permission_args=permission_args[j],
            )
        elif type(permission_args[j]) == str:
            # [DOC] Check for variables
            if permission_args[j] == '$__user':
                permission_args[j] = user['_id']
            elif permission_args[j].startswith('$__user.'):
                try:
                    permission_args[j] = _extract_attr(
                        scope=user,
                        attr_path=permission_args[j].replace('$__user.', '$__'),
                    )
                except Exception as e:
                    # [TODO] Log exception
                    # [DOC] For values that are expected to have a list value, return empty list
                    if type(permission_args) == dict and j in ['$in', '$nin']:
                        permission_args[j] = [None]
                    else:
                        permission_args[j] = None
            elif permission_args[j] == '$__access':
                permission_args[j] = {
                    '$__user': user['_id'],
                    '$__groups': user['groups'],
                }
            elif permission_args[j] == '$__datetime':
                permission_args[j] = datetime.datetime.utcnow().isoformat()
            elif permission_args[j] == '$__date':
                permission_args[j] = datetime.date.today().isoformat()
            elif permission_args[j] == '$__time':
                permission_args[j] = datetime.datetime.now().time().isoformat()

    return permission_args



import logging
from typing import Any, Dict, List, Literal, MutableMapping, Union, cast

from nawah.classes import ATTR

logger = logging.getLogger('nawah')


def _extract_attr(*, scope: MutableMapping[str, Any], attr_path: str):
    if attr_path.startswith('$__'):
        attr_path_parts = attr_path[3:].split('.')
    else:
        attr_path_parts = attr_path.split('.')
    attr: Any = scope
    for i in range(len(attr_path_parts)):
        child_attr = attr_path_parts[i]
        try:
            logger.debug(f'Attempting to extract {child_attr} from {attr}.')
            if ':' in child_attr:
                child_attr_parts = child_attr.split(':')
                attr = attr[child_attr_parts[0]]
                for i in range(1, len(child_attr_parts)):
                    attr = attr[int(child_attr_parts[i])]  # type: ignore
            else:
                attr = attr[child_attr]
        except Exception as e:
            logger.error(f'Failed to extract {child_attr} from {attr}.')
            raise e
    return attr


def _set_attr(*, scope: Dict[str, Any], attr_path: str, value: Any):
    if attr_path.startswith('$__'):
        attr_path_parts = attr_path[3:].split('.')
    else:
        attr_path_parts = attr_path.split('.')
    attr = scope
    for i in range(len(attr_path_parts) - 1):
        child_attr = attr_path_parts[i]
        try:
            if ':' in child_attr:
                child_attr_parts = child_attr.split(':')
                attr = attr[child_attr_parts[0]]
                for i in range(1, len(child_attr_parts)):
                    attr = attr[int(child_attr_parts[i])]  # type: ignore
            else:
                attr = attr[child_attr]
        except Exception as e:
            logger.error(f'Failed to extract {child_attr} from {attr}.')
            raise e
    if ':' in attr_path_parts[-1]:
        attr_path_parts_last = attr_path_parts[-1].split(':')
        attr = attr[attr_path_parts_last[0]]
        for i in range(1, len(attr_path_parts_last) - 1):
            attr = attr[int(attr_path_parts_last[i])]  # type: ignore
        attr[int(attr_path_parts_last[-1])] = value  # type: ignore
    else:
        attr[attr_path_parts[-1]] = value


def _expand_attr(*, doc: Dict[str, Any], expanded_doc: Dict[str, Any] = None):
    if not expanded_doc:
        expanded_doc = {}
    for attr in doc.keys():
        if type(doc[attr]) == dict:
            doc[attr] = _expand_attr(doc=doc[attr])
        if '.' in attr:
            attr_path = attr.split('.')
            scope = expanded_doc
            for i in range(len(attr_path) - 1):
                try:
                    if type(scope[attr_path[i]]) != dict:
                        scope[attr_path[i]] = {}
                except KeyError:
                    scope[attr_path[i]] = {}
                scope = scope[attr_path[i]]
            scope[attr_path[-1]] = doc[attr]
        else:
            expanded_doc[attr] = doc[attr]
    return expanded_doc


def _deep_update(*, target: Union[List, Dict], new_values: Union[List, Dict]):
    if type(target) != type(new_values):
        logger.error(
            f'Type \'{type(target)}\' of \'target\' is not the same as \'{type(new_values)}\' of \'new_values\'. Exiting.'
        )
        exit(1)
    if type(new_values) == dict:
        new_values = cast(dict, new_values)
        for k in new_values.keys():
            target = cast(dict, target)
            if k not in target.keys():
                target[k] = new_values[k]
            else:
                _deep_update(target=target[k], new_values=new_values[k])
    elif type(new_values) == list:
        for j in new_values:
            target = cast(list, target)
            if j not in target:
                target.append(j)


def _update_attr_values(
    *, attr: ATTR, value: Literal['default', 'extn'], value_path: str, value_val: Any
):

    value_path_part = value_path.split('.')
    for child_default_path in value_path_part:
        if ':' in child_default_path:
            if child_default_path.split(':')[0] == 'val' and attr._type == 'KV_DICT':
                attr = attr._args['val']._args['list'][
                    int(child_default_path.split(':')[1])
                ]
            else:
                attr = attr._args['dict'][child_default_path.split(':')[0]]._args[
                    'list'
                ][int(child_default_path.split(':')[1])]
        else:
            if child_default_path == 'val' and attr._type == 'KV_DICT':
                attr = attr._args['val']
            else:
                attr = attr._args['dict'][child_default_path]
    setattr(attr, f'_{value}', value_val)



from nawah.config import Config
from nawah.enums import DELETE_STRATEGY
from nawah.classes import NAWAH_ENV, ATTR, UnknownDeleteStrategyException

from bson import ObjectId
from typing import Dict, List, Any, Union

import logging

logger = logging.getLogger('nawah')


async def delete(
    *,
    env: NAWAH_ENV,
    collection_name: str,
    attrs: Dict[str, ATTR],
    docs: List[Union[str, ObjectId]],
    strategy: DELETE_STRATEGY,
) -> Dict[str, Any]:
    # [DOC] Check strategy to cherrypick update, delete calls and system_docs
    if strategy in [DELETE_STRATEGY.SOFT_SKIP_SYS, DELETE_STRATEGY.SOFT_SYS]:
        if strategy == DELETE_STRATEGY.SOFT_SKIP_SYS:
            del_docs = [
                ObjectId(doc)
                for doc in docs
                if ObjectId(doc) not in Config._sys_docs.keys()
            ]
            if len(del_docs) != len(docs):
                logger.warning(
                    'Skipped soft delete for system docs due to \'DELETE_SOFT_SKIP_SYS\' strategy.'
                )
        else:
            logger.warning('Detected \'DELETE_SOFT_SYS\' strategy for delete call.')
            del_docs = [ObjectId(doc) for doc in docs]
        # [DOC] Perform update call on matching docs
        collection = env['conn'][Config.data_name][collection_name]
        update_doc = {'$set': {'__deleted': True}}
        # [DOC] If using Azure Mongo service update docs one by one
        if Config.data_azure_mongo:
            update_count = 0
            for _id in docs:
                results = await collection.update_one({'_id': _id}, update_doc)
                update_count += results.modified_count
        else:
            results = await collection.update_many({'_id': {'$in': docs}}, update_doc)
            update_count = results.modified_count
        return {'count': update_count, 'docs': [{'_id': doc} for doc in docs]}
    elif strategy in [DELETE_STRATEGY.FORCE_SKIP_SYS, DELETE_STRATEGY.FORCE_SYS]:
        if strategy == DELETE_STRATEGY.FORCE_SKIP_SYS:
            del_docs = [
                ObjectId(doc)
                for doc in docs
                if ObjectId(doc) not in Config._sys_docs.keys()
            ]
            if len(del_docs) != len(docs):
                logger.warning(
                    'Skipped soft delete for system docs due to \'DELETE_FORCE_SKIP_SYS\' strategy.'
                )
        else:
            logger.warning('Detected \'DELETE_FORCE_SYS\' strategy for delete call.')
            del_docs = [ObjectId(doc) for doc in docs]
        # [DOC] Perform delete query on matching docs
        collection = env['conn'][Config.data_name][collection_name]
        if Config.data_azure_mongo:
            delete_count = 0
            for _id in del_docs:
                results = await collection.delete_one({'_id': _id})
                delete_count += results.deleted_count
        else:
            results = await collection.delete_many({'_id': {'$in': del_docs}})
            delete_count = results.deleted_count
        return {'count': delete_count, 'docs': [{'_id': doc} for doc in docs]}
    else:
        raise UnknownDeleteStrategyException(
            f'DELETE_STRATEGY \'{strategy}\' is unknown.'
        )
